import unittest

import uqbar.containers


class TestCase(unittest.TestCase):
    def test___init___01(self):
        """
        Initialize without name.
        """
        node = uqbar.containers.UniqueTreeList()
        assert node.name is None
        assert node.parent is None
        assert node.depth == 0
        assert len(node) == 0

    def test___init___02(self):
        """
        Initialize with name.
        """
        node = uqbar.containers.UniqueTreeList(name="foo")
        assert node.name == "foo"
        assert node.parent is None
        assert node.depth == 0
        assert len(node) == 0

    def test_append_01(self):
        """
        Unique parentage.
        """
        node_a = uqbar.containers.UniqueTreeList()
        node_b = uqbar.containers.UniqueTreeList()
        node_c = uqbar.containers.UniqueTreeNode()
        node_a.append(node_c)
        assert node_c.parent is node_a
        assert node_c in node_a
        assert len(node_a) == 1
        node_b.append(node_c)
        assert node_c.parent is node_b
        assert node_c in node_b
        assert len(node_b) == 1
        assert node_c not in node_a
        assert len(node_a) == 0

    def test_append_02(self):
        """
        Re-insertion.
        """
        node_a = uqbar.containers.UniqueTreeList()
        node_b = uqbar.containers.UniqueTreeNode()
        node_c = uqbar.containers.UniqueTreeNode()
        node_a.append(node_b)
        node_a.append(node_c)
        assert node_a is node_b.parent
        assert len(node_a) == 2
        assert node_a[:] == [node_b, node_c]
        assert node_b in node_a
        assert node_c in node_a
        node_a.append(node_b)
        assert node_a is node_b.parent
        assert len(node_a) == 2
        assert node_a[:] == [node_c, node_b]
        assert node_b in node_a
        assert node_c in node_a

    def test_extend_01(self):
        """
        Extend from a list.
        """
        node_a = uqbar.containers.UniqueTreeList()
        node_b = uqbar.containers.UniqueTreeNode()
        node_c = uqbar.containers.UniqueTreeNode()
        node_a.extend([node_b, node_c])
        assert len(node_a) == 2
        assert node_a is node_b.parent
        assert node_a is node_c.parent
        assert node_a[:] == [node_b, node_c]
        assert node_b in node_a
        assert node_c in node_a

    def test_extend_02(self):
        """
        Extend from another UniqueTreeList.
        """
        node_a = uqbar.containers.UniqueTreeList()
        node_b = uqbar.containers.UniqueTreeList()
        node_c = uqbar.containers.UniqueTreeNode()
        node_d = uqbar.containers.UniqueTreeNode()
        node_a.extend([node_c, node_d])
        assert len(node_a) == 2
        assert node_a is node_c.parent
        assert node_a is node_d.parent
        assert node_a[:] == [node_c, node_d]
        assert node_c in node_a
        assert node_d in node_a
        node_b.extend(node_a)
        assert len(node_a) == 0
        assert len(node_b) == 2
        assert node_a[:] == []
        assert node_b is node_c.parent
        assert node_b is node_d.parent
        assert node_b[:] == [node_c, node_d]
        assert node_c in node_b
        assert node_c not in node_a
        assert node_d in node_b
        assert node_d not in node_a

    def test_forbid_circular_parentage(self):
        """
        Cannot insert a parent as a child.
        """
        node_a = uqbar.containers.UniqueTreeList()
        node_b = uqbar.containers.UniqueTreeList()
        node_c = uqbar.containers.UniqueTreeList()
        node_a.append(node_b)
        node_b.append(node_c)
        with self.assertRaises(ValueError):
            node_c.append(node_a)

    def test___getitem___numbered(self):
        """
        UniqueTreeList can be subscripted with integers.
        """
        node_a = uqbar.containers.UniqueTreeList()
        node_b = uqbar.containers.UniqueTreeNode()
        node_c = uqbar.containers.UniqueTreeNode()
        node_d = uqbar.containers.UniqueTreeNode()
        node_a.extend([node_b, node_c, node_d])
        assert node_a[0] is node_b
        assert node_a[1] is node_c
        assert node_a[2] is node_d
        with self.assertRaises(IndexError):
            node_a[3]
        assert node_a[-1] is node_d
        assert node_a[-2] is node_c
        assert node_a[-3] is node_b
        with self.assertRaises(IndexError):
            node_a[-4]

    def test___getitem___named(self):
        """
        UniqueTreeList can be strings.

        Nodes are returned regardless how deep they are.
        """
        node_a = uqbar.containers.UniqueTreeList(name="foo")
        node_b = uqbar.containers.UniqueTreeList(name="bar")
        node_c = uqbar.containers.UniqueTreeNode(name="baz")
        node_d = uqbar.containers.UniqueTreeNode(name="quux")
        node_e = uqbar.containers.UniqueTreeNode(name="quux")
        node_a.extend([node_b, node_e])
        node_b.extend([node_c, node_d])
        with self.assertRaises(KeyError):
            node_a["foo"]
        assert node_a["bar"] is node_b
        assert node_a["baz"] is node_c
        assert node_a["quux"] == [node_d, node_e]

    def test___iter__(self):
        """
        UniqueTreeList can be iterated.
        """
        node_a = uqbar.containers.UniqueTreeList()
        node_b = uqbar.containers.UniqueTreeNode()
        node_c = uqbar.containers.UniqueTreeNode()
        node_d = uqbar.containers.UniqueTreeNode()
        node_a.extend([node_b, node_c, node_d])
        iterator = iter(node_a)
        assert next(iterator) is node_b
        assert next(iterator) is node_c
        assert next(iterator) is node_d
        with self.assertRaises(StopIteration):
            next(iterator)

    def test___len__(self):
        """
        Length-checking works.
        """
        node_a = uqbar.containers.UniqueTreeList()
        node_b = uqbar.containers.UniqueTreeNode()
        node_c = uqbar.containers.UniqueTreeNode()
        node_d = uqbar.containers.UniqueTreeNode()
        assert len(node_a) == 0
        node_a.append(node_b)
        assert len(node_a) == 1
        node_a.append(node_c)
        assert len(node_a) == 2
        node_a.append(node_d)
        assert len(node_a) == 3
        node_a.remove(node_b)
        assert len(node_a) == 2
        node_a.remove(node_c)
        assert len(node_a) == 1
        node_a.remove(node_d)
        assert len(node_a) == 0

    def test___setitem__(self):
        """
        Nodes can be set.
        """
        node_a = uqbar.containers.UniqueTreeList()
        node_b = uqbar.containers.UniqueTreeNode()
        node_c = uqbar.containers.UniqueTreeNode()
        node_d = uqbar.containers.UniqueTreeNode()
        node_a.extend([node_b, node_c])
        assert node_a[:] == [node_b, node_c]
        node_a[0] = node_d
        assert node_a[:] == [node_d, node_c]
        node_a[1:] = [node_b]
        assert node_a[:] == [node_d, node_b]
        node_a[:] = [node_c]
        assert node_a[:] == [node_c]
        node_a[:] = []
        assert node_a[:] == []

    def test___delitem___01(self):
        """
        Nodes can be deleted.
        """
        node_a = uqbar.containers.UniqueTreeList()
        node_b = uqbar.containers.UniqueTreeNode(name="foo")
        node_c = uqbar.containers.UniqueTreeNode(name="bar")
        node_d = uqbar.containers.UniqueTreeNode(name="baz")
        node_a.extend([node_b, node_c, node_d])
        del node_a[1]
        assert node_a[:] == [node_b, node_d]
        del node_a["baz"]
        assert node_a[:] == [node_b]

    def test___delitem___02(self):
        """
        Discontiguous named deletion works.
        """
        node_a = uqbar.containers.UniqueTreeList(name="foo")
        node_b = uqbar.containers.UniqueTreeList(name="bar")
        node_c = uqbar.containers.UniqueTreeNode(name="baz")
        node_d = uqbar.containers.UniqueTreeNode(name="quux")
        node_e = uqbar.containers.UniqueTreeNode(name="quux")
        node_a.extend([node_b, node_e])
        node_b.extend([node_c, node_d])
        del node_a["quux"]
        assert node_d.parent is None
        assert node_e.parent is None

    def test_pop(self):
        """
        Nodes can be popped.
        """
        node_a = uqbar.containers.UniqueTreeList()
        node_b = uqbar.containers.UniqueTreeNode()
        node_c = uqbar.containers.UniqueTreeNode()
        node_d = uqbar.containers.UniqueTreeNode()
        node_a.extend([node_b, node_c, node_d])
        assert node_a[:] == [node_b, node_c, node_d]
        assert node_a.pop() is node_d
        assert node_a[:] == [node_b, node_c]
        assert node_a.pop(0) is node_b
        assert node_a[:] == [node_c]
        assert node_a.pop(-1) is node_c
        assert node_a[:] == []
        assert node_b.parent is None
        assert node_c.parent is None
        assert node_d.parent is None

    def test_insert(self):
        """
        Nodes can be inserted.
        """
        node_a = uqbar.containers.UniqueTreeList()
        node_b = uqbar.containers.UniqueTreeNode()
        node_c = uqbar.containers.UniqueTreeNode()
        node_d = uqbar.containers.UniqueTreeNode()
        node_a.extend([node_b, node_c])
        assert node_a.insert(1, node_d) is None
        assert node_a is node_d.parent
        assert node_a[:] == [node_b, node_d, node_c]

    def test_remove(self):
        """
        Nodes can be removed.
        """
        node_a = uqbar.containers.UniqueTreeList()
        node_b = uqbar.containers.UniqueTreeNode()
        node_c = uqbar.containers.UniqueTreeNode()
        node_d = uqbar.containers.UniqueTreeNode()
        node_a.extend([node_b, node_c, node_d])
        assert node_a.remove(node_c) is None
        assert node_a[:] == [node_b, node_d]
        assert node_c not in node_a
        assert node_c.parent is None



import collections.abc
import enum
import math
import re
from typing import Any, FrozenSet, Mapping, Optional, Union


class Attributes(collections.abc.MutableMapping):
    """
    An attributes listing for a Graphviz graph, cluster, node or edge.

    ::

        >>> import uqbar.graphs
        >>> attributes = uqbar.graphs.Attributes(
        ...     'node',
        ...     color='black',
        ...     fillcolor='white',
        ...     shape='Mrecord',
        ...     style=['rounded', 'filled'],
        ...     )

    ::

        >>> print(format(attributes, 'graphviz'))
        [color=black,
            fillcolor=white,
            shape=Mrecord,
            style="rounded, filled"];

    """

    ### CLASS VARIABLES ###

    class Color(object):
        __slots__ = ("color",)

        def __init__(self, color) -> None:
            self.color = str(color)

        def __eq__(self, other) -> bool:
            return isinstance(other, type(self)) and self.color == other.color

        def __repr__(self) -> str:
            return "<Color {!r}>".format(self.color)

    class Mode(enum.Enum):
        CLUSTER = 1
        EDGE = 2
        GRAPH = 3
        NODE = 4
        TABLE = 5
        TABLE_CELL = 6

    class Point(object):
        __slots__ = ("x", "y")

        def __init__(self, x, y) -> None:
            self.x = float(x)
            self.y = float(y)

        def __eq__(self, other) -> bool:
            return (
                isinstance(other, type(self))
                and self.x == other.x
                and self.y == other.y
            )

    __documentation_section__ = "Core Classes"

    _arrow_types = frozenset(
        [
            "box",
            "circle",
            "crow",
            "diamond",
            "dot",
            "ediamond",
            "empty",
            "halfopen",
            "inv",
            "invdot",
            "invempty",
            "invodot",
            "none",
            "normal",
            "obox",
            "odiamond",
            "odot",
            "open",
            "tee",
            "vee",
        ]
    )

    _cluster_modes = frozenset(["global", "local", "none"])

    _dir_types = frozenset(["back", "both", "forward", "none"])

    _output_modes = frozenset(["breadthfirst", "nodesfirst", "edgesfirst"])

    _pack_modes = frozenset(["node", "clust", "graph"])

    _page_dirs = frozenset(["BL", "BR", "LB", "LT", "RB", "RT", "TL", "TR"])

    _quad_types = frozenset(["fast", "none", "normal"])

    _rank_types = frozenset(["max", "min", "same", "sink", "source"])

    _rank_dirs = frozenset(["BT", "LR", "RL", "TB"])

    _shapes = frozenset(
        [
            "Mcircle",
            "Mdiamond",
            "Msquare",
            "assembly",
            "box",
            "box3d",
            "cds",
            "circle",
            "component",
            "cylinder",
            "diamond",
            "doublecircle",
            "doubleoctagon",
            "egg",
            "ellipse",
            "fivepoverhang",
            "folder",
            "hexagon",
            "house",
            "insulator",
            "invhouse",
            "invtrapezium",
            "invtriangle",
            "larrow",
            "lpromoter",
            "none",
            "note",
            "noverhang",
            "octagon",
            "oval",
            "parallelogram",
            "pentagon",
            "plain",
            "plaintext",
            "point",
            "polygon",
            "primersite",
            "promoter",
            "proteasesite",
            "proteinstab",
            "rarrow",
            "rect",
            "rectangle",
            "restrictionsite",
            "ribosite",
            "rnastab",
            "rpromoter",
            "septagon",
            "signature",
            "square",
            "star",
            "tab",
            "terminator",
            "threepoverhang",
            "trapezium",
            "triangle",
            "tripleoctagon",
            "underline",
            "utr",
            "record",
            "Mrecord",
        ]
    )

    _smooth_types = frozenset(
        ["avg_dist", "graph_dist", "none", "power_dist", "rng", "spring", "triangle"]
    )

    _styles: FrozenSet[str] = frozenset()

    _word_pattern = re.compile(r"^\w+$")

    ### GRAPH OBJECT SPECIFICS ###

    _cluster_attributes = frozenset(
        [
            "K",
            "URL",
            "area",
            "bgcolor",
            "color",
            "colorscheme",
            "fillcolor",
            "fontcolor",
            "fontname",
            "fontsize",
            "gradientangle",
            "href",
            "id",
            "label",
            "labeljust",
            "labelloc",
            "layer",
            "lheight",
            "lp",
            "lwidth",
            "margin",
            "nojustify",
            "pencolor",
            "penwidth",
            "peripheries",
            "rank",
            "sortv",
            "style",
            "target",
            "tooltip",
        ]
    )

    _cluster_styles = frozenset(
        ["bold", "dashed", "dotted", "filled", "rounded", "solid", "striped"]
    )

    _edge_attributes = frozenset(
        [
            "arrowhead",
            "arrowsize",
            "arrowtail",
            "color",
            "colorscheme",
            "comment",
            "constraint",
            "decorate",
            "dir",
            "edgeURL",
            "edgehref",
            "edgetarget",
            "edgetooltip",
            "fillcolor",
            "fontcolor",
            "fontname",
            "fontsize",
            "headURL",
            "head_lp",
            "headclip",
            "headhref",
            "headlabel",
            "headport",
            "headtarget",
            "headtooltip",
            "href",
            "id",
            "label",
            "labelURL",
            "labelangle",
            "labeldistance",
            "labelfloat",
            "labelfontcolor",
            "labelfontname",
            "labelfontsize",
            "labelhref",
            "labeltarget",
            "labeltooltip",
            "layer",
            "len",
            "lhead",
            "lp",
            "ltail",
            "minlen",
            "nojustify",
            "penwidth",
            "pos",
            "samehead",
            "sametail",
            "showboxes",
            "style",
            "tailURL",
            "tail_lp",
            "tailclip",
            "tailhref",
            "taillabel",
            "tailport",
            "tailtarget",
            "tailtooltip",
            "target",
            "tooltip",
            "weight",
        ]
    )

    _edge_styles = frozenset(["bold", "dashed", "dotted", "solid"])

    _graph_attributes = frozenset(
        [
            "Damping",
            "K",
            "URL",
            "bb",
            "bgcolor",
            "center",
            "charset",
            "clusterrank",
            "color",
            "colorscheme",
            "comment",
            "compound",
            "concentrate",
            "defaultdist",
            "dim",
            "dimen",
            "diredgeconstraints",
            "dpi",
            "epsilon",
            "esep",
            "fontcolor",
            "fontname",
            "fontnames",
            "fontpath",
            "fontsize",
            "forcedlabels",
            "gradientangle",
            "href",
            "id",
            "imagepath",
            "inputscale",
            "label",
            "label_scheme",
            "labeljust",
            "labelloc",
            "landscape",
            "layerlistsep",
            "layers",
            "layerselect",
            "layersep",
            "layout",
            "levels",
            "levelsgap",
            "lheight",
            "lp",
            "lwidth",
            "margin",
            "maxiter",
            "mclimit",
            "mode",
            "model",
            "mosek",
            "newrank",
            "nodesep",
            "nojustify",
            "normalize",
            "notranslate",
            "nslimit",
            "nslimit1",
            "ordering",
            "orientation",
            "outputorder",
            "overlap",
            "overlap_scaling",
            "overlap_shrink",
            "pack",
            "packmode",
            "pad",
            "page",
            "pagedir",
            "quadtree",
            "quantum",
            "rank",
            "rankdir",
            "ranksep",
            "remincross",
            "repulsiveforce",
            "resolution",
            "root",
            "rotate",
            "rotation",
            "scale",
            "searchsize",
            "sep",
            "showboxes",
            "size",
            "smoothing",
            "sortv",
            "splines",
            "start",
            "style",
            "stylesheet",
            "target",
            "truecolor",
            "viewport",
            "voro_margin",
            "xdotversion",
            "xlabel",
            "xlp",
            "penwidth",
        ]
    )

    _graph_styles = _cluster_styles

    _node_attributes = frozenset(
        [
            "URL",
            "area",
            "color",
            "colorscheme",
            "comment",
            "distortion",
            "fillcolor",
            "fixedsize",
            "fontcolor",
            "fontname",
            "fontsize",
            "gradientangle",
            "group",
            "height",
            "href",
            "id",
            "image",
            "imagepos",
            "imagescale",
            "label",
            "labelloc",
            "layer",
            "margin",
            "nojustify",
            "ordering",
            "orientation",
            "penwidth",
            "peripheries",
            "pin",
            "pos",
            "rects",
            "regular",
            "root",
            "samplepoints",
            "shape",
            "shapefile",
            "showboxes",
            "sides",
            "skew",
            "sortv",
            "style",
            "target",
            "tooltip",
            "vertixes",
            "width",
            "z",
        ]
    )

    _node_styles = frozenset(
        [
            "solid",
            "dashed",
            "dotted",
            "bold",
            "rounded",
            "diagonals",
            "filled",
            "striped",
            "wedged",
        ]
    )

    ### HTML OBJECT SPECIFICS ###

    _table_attributes = frozenset(
        [
            "align",
            "bgcolor",
            "border",
            "cellborder",
            "cellpadding",
            "cellspacing",
            "color",
            "columns",
            "fixedsize",
            "gradientangle",
            "height",
            "href",
            "id",
            "rows",
            "sides",
            "style",
            "target",
            "title",
            "tooltip",
            "valign",
            "width",
        ]
    )

    _table_cell_attributes = frozenset(
        [
            "align",
            "balign",
            "bgcolor",
            "border",
            "cellpadding",
            "cellspacing",
            "color",
            "colspan",
            "fixedsize",
            "gradientangle",
            "height",
            "href",
            "id",
            "rowspan",
            "sides",
            "style",
            "target",
            "title",
            "tooltip",
            "valign",
            "width",
        ]
    )

    ### VALIDATORS ###

    _validators: Optional[Mapping[str, object]] = None

    ### INITIALIZER ###

    def __init__(self, mode: Union[str, "Attributes.Mode"], **kwargs) -> None:
        if not isinstance(mode, self.Mode):
            mode = self.Mode[str(mode).upper()]
        self._mode = mode
        self._attributes = self._validate_attributes(mode, **kwargs)

    ### SPECIAL METHODS ###

    def __delitem__(self, key: str) -> None:
        del self._attributes[key]

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, type(self))
            and self._mode == other._mode
            and self._attributes == other._attributes
        )

    def __format__(self, format_spec: Optional[str] = None) -> str:
        if format_spec == "graphviz":
            return self.__format_graphviz__()
        elif format_spec == "html":
            return self.__format_html__()
        return str(self)

    def __format_graphviz__(self) -> str:
        if not self._attributes:
            return ""
        result = []
        attributes = sorted(self._attributes.items())
        for i, (key, value) in enumerate(attributes, 1):
            value = self._format_value(value).split("\n")
            value[0] = "{}={}".format(key, value[0])
            if i < len(attributes):
                value[-1] += ","
            if i == 1:
                result.append(value.pop(0))
            result.extend("    " + _ for _ in value)
        result[0] = "[" + result[0]
        result[-1] = result[-1] + "];"
        return "\n".join(result)

    def __format_html__(self) -> str:
        if not self._attributes:
            return ""
        result = []
        for key, value in sorted(self._attributes.items()):
            value = self._format_value(value)
            if not value.startswith('"'):
                value = '"{}"'.format(value)
            result.append("{}={}".format(key.upper(), value))
        return " ".join(result)

    def __getitem__(self, key) -> Any:
        return self._attributes[key]

    def __iter__(self):
        return iter(self._attributes)

    def __len__(self):
        return len(self._attributes)

    def __setitem__(self, key, value):
        new_attributes = self._validate_attributes(self.mode, **{key: value})
        self._attributes.update(new_attributes)

    ### PRIVATE METHODS ###

    @classmethod
    def _format_value(cls, value) -> str:
        if isinstance(value, bool):
            return str(value).lower()
        elif isinstance(value, int):
            return str(value)
        elif isinstance(value, float):
            if not value % 1.0:
                value = math.floor(value)
            return str(value)
        elif isinstance(value, (list, tuple)):
            value = ", ".join(cls._format_value(x) for x in value)
            return cls._format_value(value)
        elif isinstance(value, cls.Color):
            return str(value.color)
        elif isinstance(value, str):
            if value.startswith("<") and value.endswith(">"):
                return value
            should_quote = False
            if not cls._word_pattern.match(value):
                should_quote = True
            elif value and value[0].isdigit():
                should_quote = True
            elif value.lower() in ("digraph", "edge", "graph", "node", "subgraph"):
                should_quote = True
            if should_quote:
                value = value.replace('"', r"\"")
                value = '"{}"'.format(value)
            return value
        raise ValueError(value)

    @classmethod
    def _validate_arrow_type(cls, value, **kwargs):
        value = str(value)
        assert value in cls._arrow_types
        return value

    @classmethod
    def _validate_attributes(cls, mode, **kwargs):
        valid_attributes, valid_styles = {
            cls.Mode.CLUSTER: (cls._cluster_attributes, cls._cluster_styles),
            cls.Mode.EDGE: (cls._edge_attributes, cls._edge_styles),
            cls.Mode.GRAPH: (cls._graph_attributes, cls._graph_styles),
            cls.Mode.NODE: (cls._node_attributes, cls._node_styles),
            cls.Mode.TABLE: (cls._table_attributes, ()),
            cls.Mode.TABLE_CELL: (cls._table_cell_attributes, ()),
        }[mode]
        attributes = {}
        for key, value in kwargs.items():
            if key not in valid_attributes:
                raise ValueError(key)
            validators = cls._get_validators(mode)[key]
            if not isinstance(validators, tuple):
                validators = (validators,)
            for validator in validators:
                if isinstance(validator, str):
                    if str(value) == validator:
                        value = str(value)
                        break
                    continue
                elif isinstance(validator, type):
                    value = validator(value)
                    break
                else:
                    value = validator(value, valid_styles=valid_styles)
                    break
            attributes[key] = value
        return attributes

    @classmethod
    def _validate_cluster_mode(cls, value, **kwargs):
        value = str(value)
        assert value in cls._cluster_modes
        return value

    @classmethod
    def _validate_color(cls, value, **kwargs):
        if isinstance(value, cls.Color):
            return value
        value = cls.Color(value)
        return value

    @classmethod
    def _validate_colors(cls, value, **kwargs):
        if isinstance(value, (cls.Color, int, str)):
            return cls._validate_color(value, **kwargs)
        assert len(value)
        value = tuple(cls._validate_color(_, **kwargs) for _ in value)
        if len(value) == 1:
            return value[0]
        return value

    @classmethod
    def _validate_dir_type(cls, value, **kwargs):
        value = str(value)
        assert value in cls._dir_types
        return value

    @classmethod
    def _validate_floats(cls, value, **kwargs):
        assert len(value)
        return tuple(float(_) for _ in value)

    @classmethod
    def _validate_output_mode(cls, value, **kwargs):
        value = str(value)
        assert value in cls._output_modes
        return value

    @classmethod
    def _validate_pack_mode(cls, value, **kwargs):
        value = str(value)
        assert value in cls._pack_modes
        return value

    @classmethod
    def _validate_page_dir(cls, value, **kwargs):
        value = str(value)
        assert value in cls._page_dirs
        return value

    @classmethod
    def _validate_point(cls, value, **kwargs):
        if isinstance(value, cls.Point):
            return value
        value = cls.Point(*value)
        return value

    @classmethod
    def _validate_points(cls, value_list, **kwargs):
        assert value_list
        return tuple(cls._validate_point(_, **kwargs) for _ in value_list)

    @classmethod
    def _validate_quad_type(cls, value, **kwargs):
        value = str(value)
        assert value in cls._quad_types
        return value

    @classmethod
    def _validate_rank_dir(cls, value, **kwargs):
        value = str(value)
        assert value in cls._rank_dirs
        return value

    @classmethod
    def _validate_rank_type(cls, value, **kwargs):
        value = str(value)
        assert value in cls._rank_types
        return value

    @classmethod
    def _validate_rect(cls, value, **kwargs):
        assert len(value) == 4
        value = tuple(float(_) for _ in value)
        return value

    @classmethod
    def _validate_shape(cls, value, **kwargs):
        value = str(value)
        assert value in cls._shapes
        return value

    @classmethod
    def _validate_smooth_type(cls, value, **kwargs):
        value = str(value)
        assert value in cls._smooth_types
        return value

    @classmethod
    def _validate_style(cls, value, valid_styles=None, **kwargs):
        value = str(value)
        assert value in valid_styles
        return value

    @classmethod
    def _validate_styles(cls, value, valid_styles=None, **kwargs):
        if isinstance(value, str):
            return cls._validate_style(value, valid_styles=valid_styles, **kwargs)
        assert value
        return tuple(
            cls._validate_style(_, valid_styles=valid_styles, **kwargs) for _ in value
        )

    ### PUBLIC METHODS ###

    def copy(self):
        return type(self)(self.mode, **self._attributes.copy())

    ### PUBLIC PROPERTIES ###

    @property
    def mode(self):
        return self._mode

    @classmethod
    def _get_validators(cls, mode):
        if not cls._validators:
            cls._validators = {
                "Damping": float,
                "K": float,
                "URL": str,
                "_background": str,
                "align": ("center", "left", "right", "text"),
                "area": float,
                "arrowhead": Attributes._validate_arrow_type,
                "arrowsize": float,
                "arrowtail": Attributes._validate_arrow_type,
                "balign": ("center", "left", "right", "text"),
                "bb": Attributes._validate_rect,
                "bgcolor": Attributes._validate_colors,
                "border": float,
                "cellborder": float,
                "cellpadding": float,
                "cellspacing": float,
                "center": bool,
                "charset": str,
                "clusterrank": Attributes._validate_cluster_mode,
                "color": Attributes._validate_colors,
                "colorscheme": str,
                "colspan": int,
                "columns": int,
                "comment": str,
                "compound": bool,
                "concentrate": bool,
                "constraint": bool,
                "decorate": bool,
                "defaultdist": float,
                "dim": int,
                "dimen": int,
                "dir": Attributes._validate_dir_type,
                "diredgeconstraints": ("hier", bool),
                "distortion": float,
                "dpi": float,
                "edgeURL": str,
                "edgehref": str,
                "edgetarget": str,
                "edgetooltip": str,
                "epsilon": float,
                "esep": (float, Attributes._validate_point),
                "fillcolor": Attributes._validate_colors,
                "fixedsize": ("shape", bool),
                "fontcolor": Attributes._validate_color,
                "fontname": str,
                "fontnames": str,
                "fontpath": str,
                "fontsize": float,
                "forcelabels": bool,
                "gradientangle": int,
                "group": str,
                "headURL": str,
                "head_lp": Attributes._validate_point,
                "headclip": bool,
                "headhref": str,
                "headlabel": str,
                "headport": str,
                "headtarget": str,
                "headtooltip": str,
                "height": float,
                "href": str,
                "id": str,
                "image": str,
                "imagepath": str,
                "imagepos": str,
                "imagescale": ("width", "height", "both", bool),
                "inputscale": float,
                "label": str,
                "labelURL": str,
                "label_scheme": int,
                "labelangle": float,
                "labeldistance": float,
                "labelfloat": bool,
                "labelfontcolor": Attributes._validate_color,
                "labelfontname": str,
                "labelfontsize": float,
                "labelhref": str,
                "labeljust": str,
                "labelloc": str,
                "labeltarget": str,
                "labeltooltip": str,
                "landscape": bool,
                "layer": str,
                "layerlistsep": str,
                "layers": str,
                "layerselect": str,
                "layersep": str,
                "layout": str,
                "len": float,
                "levels": int,
                "levelsgap": float,
                "lhead": str,
                "lheight": float,
                "lp": Attributes._validate_point,
                "ltail": str,
                "lwidth": float,
                "margin": (float, Attributes._validate_point),
                "maxiter": int,
                "mclimit": float,
                "mindist": float,
                "minlen": int,
                "mode": str,
                "model": str,
                "mosek": bool,
                "newrank": bool,
                "nodesep": float,
                "nojustify": bool,
                "normalize": (float, bool),
                "notranslate": bool,
                "nslimit": float,
                "nslimit1": float,
                "ordering": str,
                "outputorder": Attributes._validate_output_mode,
                "overlap": ("scale", "scalexy", "compress", "ipsep", "prism", bool),
                "overlap_scaling": float,
                "overlap_shrink": bool,
                "pack": bool,
                "packmode": Attributes._validate_pack_mode,
                "pad": (float, Attributes._validate_point),
                "page": (float, Attributes._validate_point),
                "pagedir": Attributes._validate_page_dir,
                "pencolor": Attributes._validate_color,
                "penwidth": float,
                "peripheries": int,
                "pin": bool,
                "pos": Attributes._validate_point,
                "quadtree": (Attributes._validate_quad_type, bool),
                "quantum": float,
                "rank": Attributes._validate_rank_type,
                "rankdir": Attributes._validate_rank_dir,
                "ranksep": (float, Attributes._validate_floats),
                "ratio": ("fill", "compress", "expand", "auto", float),
                "rects": Attributes._validate_rect,
                "regular": bool,
                "remincross": bool,
                "repulsiveforce": float,
                "resolution": float,
                "root": str,
                "rotate": int,
                "rotation": float,
                "rows": int,
                "rowspan": int,
                "samehead": str,
                "sametail": str,
                "samplepoints": int,
                "scale": (float, Attributes._validate_point),
                "searchsize": int,
                "sep": (float, Attributes._validate_point),
                "shape": Attributes._validate_shape,
                "shapefile": str,
                "showboxes": int,
                "sides": (int, str),
                "size": (float, Attributes._validate_point),
                "skew": float,
                "smoothing": Attributes._validate_smooth_type,
                "sortv": int,
                "splines": (
                    "none",
                    "line",
                    "polyline",
                    "curved",
                    "ortho",
                    "spline",
                    bool,
                ),
                "start": str,
                "style": Attributes._validate_styles,
                "stylesheet": str,
                "tailURL": str,
                "tail_lp": Attributes._validate_point,
                "tailclip": bool,
                "tailhref": str,
                "taillabel": str,
                "tailport": str,
                "tailtarget": str,
                "tailtooltip": str,
                "target": str,
                "title": str,
                "tooltip": str,
                "truecolor": bool,
                "valign": ("middle", "bottom", "top"),
                "vertices": Attributes._validate_points,
                "viewport": str,
                "voro_margin": float,
                "weight": int,
                "width": float,
                "xdotversion": str,
                "xlabel": str,
                "xlp": Attributes._validate_point,
                "z": float,
            }
        return cls._validators



from xml.dom import minidom  # type: ignore

from ..graphs import Grapher


def load_ipython_extension(ipython):
    patch_grapher()


def patch_grapher():
    def get_format(self):
        return "svg"

    def open_output_path(self, output_path):
        from IPython.core.display import display_svg

        with output_path.open() as file_pointer:
            contents = file_pointer.read()
        delete_attributes = True
        document = minidom.parseString(contents)
        svg_element = document.getElementsByTagName("svg")[0]
        view_box = svg_element.getAttribute("viewBox")
        view_box = [float(_) for _ in view_box.split()]
        if delete_attributes:
            if svg_element.attributes.get("height", None):
                del svg_element.attributes["height"]
            if svg_element.attributes.get("width", None):
                del svg_element.attributes["width"]
        else:
            height = "{}pt".format(int(view_box[-1] * 0.6))
            width = "{}pt".format(int(view_box[-2] * 0.6))
            svg_element.setAttribute("height", height)
            svg_element.setAttribute("width", width)
        svg_element.setAttribute("preserveAspectRatio", "xMinYMin")
        contents = document.toprettyxml()
        display_svg(contents, raw=True)

    Grapher.get_format = get_format
    Grapher.open_output_path = open_output_path



import sys

import pytest


@pytest.mark.sphinx("html", testroot="uqbar-sphinx-style")
def test_sphinx_style_html(app, status, warning):
    app.build()
    assert "build succeeded" in status.getvalue()
    warnings = [line.strip() for line in warning.getvalue().splitlines()]
    if sys.version_info.minor < 11:
        assert not warnings
    else:
        assert len(warnings) == 2


@pytest.mark.sphinx("latex", testroot="uqbar-sphinx-style")
def test_sphinx_style_latex(app, status, warning):
    app.build()
    assert "build succeeded" in status.getvalue()
    warnings = [line.strip() for line in warning.getvalue().splitlines()]
    assert not warnings



from typing import Mapping, Optional, Tuple, Union

from ..containers import UniqueTreeList, UniqueTreeNode
from .attrs import Attributes
from .core import Attachable


class HRule(UniqueTreeNode):
    """
    A Graphviz HTML horizontal rule.

    ::

        >>> import uqbar.graphs
        >>> hrule = uqbar.graphs.HRule()
        >>> print(format(hrule, 'graphviz'))
        <HR/>

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "HTML Classes"

    ### SPECIAL METHODS ###

    def __format__(self, format_spec: Optional[str] = None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == "graphviz":
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        return "<HR/>"


class LineBreak(UniqueTreeNode):
    """
    A Graphviz HTML line break.

    ::

        >>> import uqbar.graphs
        >>> line_break = uqbar.graphs.LineBreak()
        >>> print(format(line_break, 'graphviz'))
        <BR/>

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "HTML Classes"

    ### SPECIAL METHODS ###

    def __format__(self, format_spec: Optional[str] = None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == "graphviz":
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        return "<BR/>"


class Table(UniqueTreeList):
    """
    A Graphviz HTML table.

    ::

        >>> import uqbar.graphs
        >>> table = uqbar.graphs.Table([
        ...     uqbar.graphs.TableRow([
        ...         uqbar.graphs.TableCell('Cell One'),
        ...     ]),
        ...     uqbar.graphs.HRule(),
        ...     uqbar.graphs.TableRow([
        ...         uqbar.graphs.TableCell('Cell Two'),
        ...     ]),
        ... ])
        >>> print(format(table, 'graphviz'))
        <TABLE>
            <TR>
                <TD>Cell One</TD>
            </TR>
            <HR/>
            <TR>
                <TD>Cell Two</TD>
            </TR>
        </TABLE>

    ::

        >>> table = uqbar.graphs.Table(
        ...     attributes={
        ...         'border': 5,
        ...         'bgcolor': 'blue',
        ...         },
        ...     )
        >>> print(format(table, 'graphviz'))
        <TABLE BGCOLOR="blue" BORDER="5"></TABLE>

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "HTML Classes"

    ### INITIALIZER ###

    def __init__(
        self,
        children=None,
        *,
        attributes: Optional[Union[Mapping[str, object], Attributes]] = None,
        name: Optional[str] = None,
    ) -> None:
        UniqueTreeList.__init__(self, children=children, name=name)
        self._attributes = Attributes("table", **(attributes or {}))

    ### SPECIAL METHODS ###

    def __format__(self, format_spec: Optional[str] = None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == "graphviz":
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        result = []
        start, stop = "<TABLE", "</TABLE>"
        attributes = format(self._attributes, "html")
        if attributes:
            start += " {}".format(attributes)
        start += ">"
        result.append(start)
        for child in self:
            for line in format(child, "graphviz").splitlines():
                result.append("    {}".format(line))
        result.append(stop)
        join_character = ""
        if len(result) > 2:
            join_character = "\n"
        return join_character.join(result)

    ### PRIVATE PROPERTIES ###

    @property
    def _node_class(self) -> Tuple[type, ...]:
        import uqbar.graphs

        return (uqbar.graphs.TableRow, uqbar.graphs.HRule)


class TableRow(UniqueTreeList):
    """
    A Graphviz HTML table row.

    ::

        >>> import uqbar.graphs
        >>> table_row = uqbar.graphs.TableRow()
        >>> print(format(table_row, 'graphviz'))
        <TR></TR>

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "HTML Classes"

    ### SPECIAL METHODS ###

    def __format__(self, format_spec: Optional[str] = None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == "graphviz":
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        result = []
        result.append("<TR>")
        for child in self:
            for line in format(child, "graphviz").splitlines():
                result.append("    {}".format(line))
        result.append("</TR>")
        join_character = ""
        if len(result) > 2:
            join_character = "\n"
        return join_character.join(result)

    ### PRIVATE PROPERTIES ###

    @property
    def _node_class(self) -> Tuple[type, ...]:
        import uqbar.graphs

        return (uqbar.graphs.TableCell, uqbar.graphs.VRule)


class TableCell(UniqueTreeList, Attachable):
    """
    A Graphviz HTML table.

    ::

        >>> import uqbar.graphs
        >>> table_cell = uqbar.graphs.TableCell()
        >>> print(format(table_cell, 'graphviz'))
        <TD></TD>

    ::

        >>> table_cell = uqbar.graphs.TableCell(
        ...     attributes={
        ...         'border': 5,
        ...         'bgcolor': 'blue',
        ...         },
        ...     )
        >>> print(format(table_cell, 'graphviz'))
        <TD BGCOLOR="blue" BORDER="5"></TD>

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "HTML Classes"

    ### INITIALIZER ###

    def __init__(
        self,
        children=None,
        *,
        attributes: Optional[Union[Mapping[str, object], Attributes]] = None,
        name: Optional[str] = None,
    ) -> None:
        if isinstance(children, str):
            children = [Text(children)]
        UniqueTreeList.__init__(self, children=children, name=name)
        Attachable.__init__(self)
        self._attributes = Attributes("table_cell", **(attributes or {}))

    ### SPECIAL METHODS ###

    def __format__(self, format_spec: Optional[str] = None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == "graphviz":
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        result = []
        start, stop = "<TD", "</TD>"
        if self.edges:
            start += ' PORT="{}"'.format(self._get_port_name())
        attributes = format(self._attributes, "html")
        if attributes:
            start += " {}".format(attributes)
        start += ">"
        result.append(start)
        for child in self:
            result.append(format(child, "graphviz"))
        result.append(stop)
        return "".join(result)

    ### PRIVATE PROPERTIES ###

    @property
    def _node_class(self) -> Tuple[type, ...]:
        import uqbar.graphs

        return (uqbar.graphs.Table, uqbar.graphs.LineBreak, uqbar.graphs.Text)


class Text(UniqueTreeNode):
    """
    A Graphviz HTML text node.

    ::

        >>> import uqbar.graphs
        >>> text = uqbar.graphs.Text('text')
        >>> print(format(text, 'graphviz'))
        text

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "HTML Classes"

    ### INITIALIZER ###

    def __init__(self, text, *, name=None):
        UniqueTreeNode.__init__(self, name=name)
        self.text = text

    ### SPECIAL METHODS ###

    def __format__(self, format_spec: Optional[str] = None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == "graphviz":
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        return self.text

    ### PUBLIC PROPERTIES ###

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, text):
        if text is not None:
            text = str(text)
        self._text = text


class VRule(UniqueTreeNode):
    """
    A Graphviz HTML vertical rule.

    ::

        >>> import uqbar.graphs
        >>> hrule = uqbar.graphs.VRule()
        >>> print(format(hrule, 'graphviz'))
        <VR/>

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = "HTML Classes"

    ### SPECIAL METHODS ###

    def __format__(self, format_spec: Optional[str] = None) -> str:
        # TODO: make the format specification options machine-readable
        if format_spec == "graphviz":
            return self.__format_graphviz__()
        return str(self)

    def __format_graphviz__(self) -> str:
        return "<VR/>"



import abc
import copy
import hashlib
import pathlib
import subprocess

from docutils.nodes import FixedTextElement, General, SkipNode

from ..graphs import Grapher
from .console import Console
from .sphinx import UqbarBookDefaultsDirective, UqbarBookDirective


class Extension:
    @classmethod
    def add_option(cls, key, value):
        UqbarBookDirective.option_spec[key] = value
        UqbarBookDefaultsDirective.option_spec[key] = value

    @classmethod
    @abc.abstractmethod
    def setup_console(cls, console: Console, monkeypatch):
        """
        Perform console setup tasks before executing a suite of code blocks,
        e.g. monkeypatching IO operations to allow media capture.
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def setup_sphinx(cls, app):
        """
        Setup Sphinx, e.g. register custom nodes and visitors.
        """
        raise NotImplementedError

    @classmethod
    def teardown_console(cls, app):
        """
        Perform console teardown tasks.
        """
        pass

    @staticmethod
    def visit_block_html(self, node):
        raise SkipNode

    @staticmethod
    def visit_block_latex(self, node):
        raise SkipNode

    @staticmethod
    def depart_block_text(self, node):
        self.end_state(wrap=False)

    @staticmethod
    def visit_block_text(self, node):
        self.new_state()


class GraphExtension(Extension):
    template = (
        '<div class="uqbar-book">'
        '<a href="{dot_file_path}" title="{title}">'
        '<img src="{image_file_path}" alt="{alt}"/>'
        "</a>"
        "</div>"
    )

    class graphviz_block(General, FixedTextElement):
        pass

    @classmethod
    def setup_console(cls, console, monkeypatch):
        monkeypatch.setattr(
            Grapher,
            "__call__",
            lambda self: console.push_proxy(
                cls(
                    self.graphable,
                    self.layout,
                    **{
                        key.replace("graphviz/", "").replace("-", "_"): value
                        for key, value in console.proxy_options.items()
                        if key.startswith("graphviz/")
                    },
                )
            ),
        )

    @classmethod
    def setup_sphinx(cls, app):
        app.add_node(
            cls.graphviz_block,
            html=[cls.visit_block_html, None],
            latex=[cls.visit_block_latex, None],
            text=[cls.visit_block_text, cls.depart_block_text],
        )

    def __init__(self, graphable, layout):
        try:
            self.graphable = copy.deepcopy(graphable)
        except Exception:
            self.graphable = copy.deepcopy(graphable.__graph__())
        self.layout = layout

    def to_docutils(self):
        try:
            graphviz_graph = self.graphable.__graph__()
            code = format(graphviz_graph, "graphviz")
        except AttributeError:
            code = self.graphable.__format_graphviz__()
        node = self.graphviz_block(code, code)
        node["layout"] = self.layout
        return [node]

    @classmethod
    def clean_svg(cls, svg_path):
        pass

    @classmethod
    def render_image(cls, node, output_path, suffix):
        sha256 = hashlib.sha256()
        sha256.update(node[0].encode())
        sha256.update(node["layout"].encode())
        hexdigest = sha256.hexdigest()
        base_path = output_path / "graphviz-{}".format(hexdigest)
        base_path.parent.mkdir(exist_ok=True, parents=True)
        image_file_path = base_path.with_suffix(suffix)
        dot_file_path = base_path.with_suffix(".dot")
        if image_file_path.exists() and dot_file_path.exists():
            return image_file_path
        if not dot_file_path.exists():
            dot_file_path.write_text(node[0])
        if not image_file_path.exists():
            command = "{layout} -T {format} -o {image_path} {dot_path}".format(
                layout=node["layout"],
                format=suffix.strip("."),
                image_path=image_file_path,
                dot_path=dot_file_path,
            )
            subprocess.call(command, shell=True)
            if suffix == ".svg":
                cls.clean_svg(image_file_path)
        return image_file_path

    @staticmethod
    def visit_block_html(self, node):
        absolute_image_file_path = GraphExtension.render_image(
            node, pathlib.Path(self.builder.outdir) / "_images", ".svg"
        )
        relative_image_file_path = (
            pathlib.Path(self.builder.imgpath) / absolute_image_file_path.name
        )
        result = GraphExtension.template.format(
            dot_file_path=relative_image_file_path.with_suffix(".dot"),
            image_file_path=relative_image_file_path,
            title="",
            alt="",
        )
        self.body.append(result)
        raise SkipNode



import unittest

import uqbar.containers


class TestCase(unittest.TestCase):
    def test___init___01(self):
        """
        Initialize without name.
        """
        node = uqbar.containers.UniqueTreeNode()
        assert node.name is None
        assert node.parent is None
        assert node.depth == 0

    def test___init___02(self):
        """
        Initialize with name.
        """
        node = uqbar.containers.UniqueTreeNode(name="foo")
        assert node.name == "foo"
        assert node.parent is None
        assert node.depth == 0

    def test_name_01(self):
        node = uqbar.containers.UniqueTreeNode(name="foo")
        node.name = "bar"
        assert node.name == "bar"

    def test_name_02(self):
        node_a = uqbar.containers.UniqueTreeList()
        node_b = uqbar.containers.UniqueTreeList()
        node_c = uqbar.containers.UniqueTreeNode(name="foo")
        node_d = uqbar.containers.UniqueTreeList()
        node_a.append(node_b)
        node_b.append(node_c)
        assert node_a["foo"] is node_c
        assert node_b["foo"] is node_c
        with self.assertRaises(KeyError):
            node_d["foo"]
        node_c.name = "bar"
        assert node_a["bar"] is node_c
        assert node_b["bar"] is node_c
        with self.assertRaises(KeyError):
            node_a["foo"]
        with self.assertRaises(KeyError):
            node_b["foo"]
        with self.assertRaises(KeyError):
            node_d["bar"]
        node_d.append(node_c)
        assert node_d["bar"] is node_c
        with self.assertRaises(KeyError):
            node_a["bar"]
        with self.assertRaises(KeyError):
            node_b["bar"]



import unittest

import uqbar.graphs
from uqbar.strings import normalize


class TestCase(unittest.TestCase):
    def test___init__(self):
        node = uqbar.graphs.Node()
        assert node.name is None

        node = uqbar.graphs.Node(name="foo")
        assert node.name == "foo"
        assert not node.attributes

        attributes = uqbar.graphs.Attributes(mode="node", color="blue")
        node = uqbar.graphs.Node(name="foo", attributes=attributes)
        assert node.name == "foo"
        assert len(node.attributes) == 1
        assert node.attributes["color"] == node.attributes.Color("blue")

    def test__get_canonical_name(self):
        node = uqbar.graphs.Node()
        assert node._get_canonical_name() == "node_0"

        node = uqbar.graphs.Node(name="foo")
        assert node._get_canonical_name() == "foo"

        node = uqbar.graphs.Node(name="Foo Bar Baz")
        assert node._get_canonical_name() == "Foo Bar Baz"

        graph = uqbar.graphs.Graph()
        node = uqbar.graphs.Node()
        graph.append(node)
        assert node._get_canonical_name() == "node_0"

        graph = uqbar.graphs.Graph()
        subgraph = uqbar.graphs.Graph()
        node = uqbar.graphs.Node()
        graph.append(subgraph)
        subgraph.append(node)
        assert node._get_canonical_name() == "node_0_0"

        graph = uqbar.graphs.Graph()
        node_a = uqbar.graphs.Node(name="foo")
        node_b = uqbar.graphs.Node(name="foo")
        node_c = uqbar.graphs.Node(name="bar")
        graph.extend([node_a, node_b, node_c])
        assert node_a._get_canonical_name() == "foo_0"
        assert node_b._get_canonical_name() == "foo_1"
        assert node_c._get_canonical_name() == "bar"

    def test___format___str(self):
        node = uqbar.graphs.Node()
        assert format(node) == repr(node)

        node = uqbar.graphs.Node(name="foo")
        assert format(node) == repr(node)

        attributes = uqbar.graphs.Attributes(mode="node", color="blue")
        node = uqbar.graphs.Node(name="foo", attributes=attributes)
        assert format(node) == repr(node)

    def test___format___graphviz(self):
        node = uqbar.graphs.Node()
        assert format(node, "graphviz") == "node_0;"

        node = uqbar.graphs.Node(name="foo")
        assert format(node, "graphviz") == "foo;"

        attributes = uqbar.graphs.Attributes(
            mode="node",
            color="blue",
            fontname="Times New Roman",
            fontsize=11.5,
            shape="oval",
        )
        node = uqbar.graphs.Node(name="foo", attributes=attributes)
        assert format(node, "graphviz") == normalize(
            """
            foo [color=blue,
                fontname="Times New Roman",
                fontsize=11.5,
                shape=oval];
            """
        )



import code
import inspect
import itertools
import sys
import traceback
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple

from ..io import RedirectedStreams


@dataclass(frozen=True)
class ConsoleInput:
    string: str


@dataclass(frozen=True)
class ConsoleOutput:
    string: str


class ConsoleError(Exception):
    pass


unset = object()


class MonkeyPatch(object):
    def __init__(self):
        self._attributes = []

    def setattr(self, target, name, value=unset):
        oldval = getattr(target, name, unset)
        if inspect.isclass(target):
            oldval = target.__dict__.get(name, unset)
        self._attributes.append((target, name, oldval))
        setattr(target, name, value)

    def undo(self):
        for obj, name, value in reversed(self._attributes):
            if value is not unset:
                setattr(obj, name, value)
            else:
                delattr(obj, name)
        self._attributes[:] = []


class Console(code.InteractiveConsole):
    r"""
    Interactive console providing a sandboxed namespace for executing code
    examples.

    .. container:: example

        ::

            >>> from uqbar.book.console import Console
            >>> console = Console()
            >>> results, errored = console.interpret([
            ...     "x = 3",
            ...     "for i in range(3):",
            ...     "    print(i + x)",
            ...     "",
            ...     "quit()",  # SystemExit is captured
            ...     "print(2 + 2)",
            ... ])

        ::

            >>> for result in results:
            ...     print(result)
            ...
            ConsoleInput(string='>>> x = 3\n>>> for i in range(3):\n...     print(i + x)\n... \n')
            ConsoleOutput(string='3\n4\n5\n')
            ConsoleInput(string='>>> quit()\n>>> print(2 + 2)\n')
            ConsoleOutput(string='4\n')

        ::

            >>> print(errored)
            False

    .. container:: example

        The console maintains state across multiple ``interpret()`` calls:

        ::

            >>> results, errored = console.interpret(["x + 2"])
            >>> for result in results:
            ...     print(result)
            ...
            ConsoleInput(string='>>> x + 2\n')
            ConsoleOutput(string='5\n')

        ::

            >>> print(errored)
            False

    .. container:: example

        Errors are reported but do not propagate outside of the ``interpret()`` call:

        ::

            >>> results, errored = console.interpret(["1 / 0"])
            >>> for result in results:
            ...     print(result)
            ...
            ConsoleInput(string='>>> 1 / 0\n')
            ConsoleOutput(string='Traceback (most recent call last):\n  File "<stdin>", line 1, in <module>\nZeroDivisionError: division by zero\n')

        ::

            >>> print(errored)
            True

    .. container:: example

        Error reporting resets:

        ::

            >>> results, errored = console.interpret(["x += 5", "print(x)"])
            >>> for result in results:
            ...     print(result)
            ...
            ConsoleInput(string='>>> x += 5\n>>> print(x)\n')
            ConsoleOutput(string='8\n')

        ::

            >>> print(errored)
            False

    """

    ### INITIALIZER ###

    def __init__(
        self, namespace: Optional[Dict] = None, extensions: Optional[List[Any]] = None
    ):
        super().__init__(
            filename="<stdin>",
            locals={**(namespace or {}), "__name__": "__main__", "__package__": None},
        )
        self.extensions = extensions
        self.errored = False
        self.results: List[Any] = []
        self.monkeypatch = None
        self.proxy_options: Dict[str, Dict[str, Any]] = {}

    ### SPECIAL METHODS ###

    def __enter__(self):
        self.monkeypatch = MonkeyPatch()
        for extension in self.extensions or []:
            extension.setup_console(self, self.monkeypatch)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.monkeypatch.undo()
        for extension in self.extensions or []:
            extension.teardown_console(self)

    ### PRIVATE METHODS ###

    def _showtraceback(self):
        """
        Re-implementation of code.InteractiveConsole's showtraceback().

        No new functionality, but placing it here prevents the addition of
        extra lines in the trace output when run under GHA.
        """
        sys.last_type, sys.last_value, last_tb = ei = sys.exc_info()
        sys.last_traceback = last_tb
        try:
            self.write(
                "".join(traceback.format_exception(ei[0], ei[1], last_tb.tb_next))
            )
        finally:
            last_tb = ei = None

    ### PUBLIC METHODS ###

    def flush(self) -> None:
        pass

    def interpret(self, lines: List[str]) -> Tuple[List[Any], bool]:
        self.resetbuffer()
        self.errored = False
        self.results[:] = []
        is_incomplete_statement = False
        with RedirectedStreams(self, self), self:
            prompt = ">>> "
            for line_number, line in enumerate(lines, 1):
                self.results.append(ConsoleInput(prompt + line + "\n"))
                try:
                    is_incomplete_statement = self.push(line)
                except SystemExit:
                    self.resetbuffer()
                prompt = "... " if is_incomplete_statement else ">>> "
            if is_incomplete_statement:
                self.results.append(ConsoleInput(prompt + "\n"))
                try:
                    is_incomplete_statement = self.push("\n")
                except SystemExit:
                    self.resetbuffer()
            if is_incomplete_statement:
                self.errored = True
        results = []
        for class_, grouper in itertools.groupby(self.results, lambda x: type(x)):
            if class_ is ConsoleInput:
                # Strip extra blank lines induced by black.
                string = "".join(_.string for _ in grouper if _.string.strip() != ">>>")
                results.append(class_(string=string))
            elif class_ is ConsoleOutput:
                results.append(class_(string="".join(_.string for _ in grouper)))
            else:
                results.extend(grouper)
        return results, self.errored

    def push_proxy(self, proxy):
        self.results.append(proxy)

    def push_proxy_options(self, options=None):
        self.proxy_options = options or {}

    def showsyntaxerror(self, filename: Optional[str] = None) -> None:
        super().showsyntaxerror(filename=filename)
        self.errored = True

    def showtraceback(self) -> None:
        self._showtraceback()
        self.errored = True

    def write(self, string: str) -> None:
        self.results.append(ConsoleOutput(string))



import collections
import pickle
import sys

import pytest

import uqbar.apis
import uqbar.apis.dummy
import uqbar.strings


def test_01():
    inheritance_graph = uqbar.apis.InheritanceGraph(package_paths=["uqbar.containers"])
    assert str(inheritance_graph) == uqbar.strings.normalize(
        r"""
            digraph InheritanceGraph {
                graph [bgcolor=transparent,
                    color=lightsteelblue2,
                    fontname=Arial,
                    fontsize=10,
                    outputorder=edgesfirst,
                    overlap=prism,
                    penwidth=2,
                    rankdir=LR,
                    splines=spline,
                    style="dashed, rounded",
                    truecolor=true];
                node [colorscheme=pastel19,
                    fontname=Arial,
                    fontsize=10,
                    height=0,
                    penwidth=2,
                    shape=box,
                    style="filled, rounded",
                    width=0];
                edge [color=lightslategrey,
                    penwidth=1];
                subgraph cluster_builtins {
                    graph [label=builtins];
                    node [color=1];
                    "builtins.object" [label=object];
                }
                subgraph "cluster_uqbar.containers.dependency_graph" {
                    graph [label="uqbar.containers.dependency_graph"];
                    node [color=2];
                    "uqbar.containers.dependency_graph.DependencyGraph" [label="Dependency\nGraph"];
                }
                subgraph "cluster_uqbar.containers.unique_tree" {
                    graph [label="uqbar.containers.unique_tree"];
                    node [color=3];
                    "uqbar.containers.unique_tree.UniqueTreeContainer" [label="Unique\nTree\nContainer"];
                    "uqbar.containers.unique_tree.UniqueTreeDict" [label="Unique\nTree\nDict"];
                    "uqbar.containers.unique_tree.UniqueTreeList" [label="Unique\nTree\nList"];
                    "uqbar.containers.unique_tree.UniqueTreeNode" [label="Unique\nTree\nNode"];
                    "uqbar.containers.unique_tree.UniqueTreeSet" [label="Unique\nTree\nSet"];
                    "uqbar.containers.unique_tree.UniqueTreeTuple" [label="Unique\nTree\nTuple"];
                    "uqbar.containers.unique_tree.UniqueTreeContainer" -> "uqbar.containers.unique_tree.UniqueTreeDict";
                    "uqbar.containers.unique_tree.UniqueTreeContainer" -> "uqbar.containers.unique_tree.UniqueTreeList";
                    "uqbar.containers.unique_tree.UniqueTreeContainer" -> "uqbar.containers.unique_tree.UniqueTreeSet";
                    "uqbar.containers.unique_tree.UniqueTreeContainer" -> "uqbar.containers.unique_tree.UniqueTreeTuple";
                    "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.containers.unique_tree.UniqueTreeContainer";
                }
                "builtins.object" -> "uqbar.containers.dependency_graph.DependencyGraph";
                "builtins.object" -> "uqbar.containers.unique_tree.UniqueTreeNode";
            }
        """
    )
    pickle.dumps(inheritance_graph)


def test_02():
    inheritance_graph = uqbar.apis.InheritanceGraph(
        package_paths=["uqbar"], lineage_paths=["uqbar.containers"]
    )
    assert str(inheritance_graph) == uqbar.strings.normalize(
        r"""
            digraph InheritanceGraph {
                graph [bgcolor=transparent,
                    color=lightsteelblue2,
                    fontname=Arial,
                    fontsize=10,
                    outputorder=edgesfirst,
                    overlap=prism,
                    penwidth=2,
                    rankdir=LR,
                    splines=spline,
                    style="dashed, rounded",
                    truecolor=true];
                node [colorscheme=pastel19,
                    fontname=Arial,
                    fontsize=10,
                    height=0,
                    penwidth=2,
                    shape=box,
                    style="filled, rounded",
                    width=0];
                edge [color=lightslategrey,
                    penwidth=1];
                subgraph cluster_builtins {
                    graph [label=builtins];
                    node [color=1];
                    "builtins.object" [label=object];
                }
                subgraph "cluster_uqbar.apis.nodes" {
                    graph [label="uqbar.apis.nodes"];
                    node [color=2];
                    "uqbar.apis.nodes.ModuleNode" [label="Module\nNode"];
                    "uqbar.apis.nodes.PackageNode" [label="Package\nNode"];
                }
                subgraph "cluster_uqbar.containers.dependency_graph" {
                    graph [label="uqbar.containers.dependency_graph"];
                    node [color=3];
                    "uqbar.containers.dependency_graph.DependencyGraph" [label="Dependency\nGraph"];
                }
                subgraph "cluster_uqbar.containers.unique_tree" {
                    graph [label="uqbar.containers.unique_tree"];
                    node [color=4];
                    "uqbar.containers.unique_tree.UniqueTreeContainer" [label="Unique\nTree\nContainer"];
                    "uqbar.containers.unique_tree.UniqueTreeDict" [label="Unique\nTree\nDict"];
                    "uqbar.containers.unique_tree.UniqueTreeList" [label="Unique\nTree\nList"];
                    "uqbar.containers.unique_tree.UniqueTreeNode" [label="Unique\nTree\nNode"];
                    "uqbar.containers.unique_tree.UniqueTreeSet" [label="Unique\nTree\nSet"];
                    "uqbar.containers.unique_tree.UniqueTreeTuple" [label="Unique\nTree\nTuple"];
                    "uqbar.containers.unique_tree.UniqueTreeContainer" -> "uqbar.containers.unique_tree.UniqueTreeDict";
                    "uqbar.containers.unique_tree.UniqueTreeContainer" -> "uqbar.containers.unique_tree.UniqueTreeList";
                    "uqbar.containers.unique_tree.UniqueTreeContainer" -> "uqbar.containers.unique_tree.UniqueTreeSet";
                    "uqbar.containers.unique_tree.UniqueTreeContainer" -> "uqbar.containers.unique_tree.UniqueTreeTuple";
                    "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.containers.unique_tree.UniqueTreeContainer";
                }
                subgraph "cluster_uqbar.graphs.core" {
                    graph [label="uqbar.graphs.core"];
                    node [color=5];
                    "uqbar.graphs.core.Attachable" [label=Attachable];
                    "uqbar.graphs.core.Graph" [label="Graph"];
                    "uqbar.graphs.core.Node" [label="Node"];
                }
                subgraph "cluster_uqbar.graphs.html" {
                    graph [label="uqbar.graphs.html"];
                    node [color=6];
                    "uqbar.graphs.html.HRule" [label=HRule];
                    "uqbar.graphs.html.LineBreak" [label="Line\nBreak"];
                    "uqbar.graphs.html.Table" [label=Table];
                    "uqbar.graphs.html.TableCell" [label="Table\nCell"];
                    "uqbar.graphs.html.TableRow" [label="Table\nRow"];
                    "uqbar.graphs.html.Text" [label=Text];
                    "uqbar.graphs.html.VRule" [label=VRule];
                }
                subgraph "cluster_uqbar.graphs.records" {
                    graph [label="uqbar.graphs.records"];
                    node [color=7];
                    "uqbar.graphs.records.RecordField" [label="Record\nField"];
                    "uqbar.graphs.records.RecordGroup" [label="Record\nGroup"];
                }
                "builtins.object" -> "uqbar.containers.dependency_graph.DependencyGraph";
                "builtins.object" -> "uqbar.containers.unique_tree.UniqueTreeNode";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.apis.nodes.PackageNode";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.graphs.core.Graph";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.graphs.core.Node";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.graphs.html.Table";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.graphs.html.TableCell";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.graphs.html.TableRow";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.graphs.records.RecordGroup";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.apis.nodes.ModuleNode";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.graphs.core.Attachable";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.graphs.html.HRule";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.graphs.html.LineBreak";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.graphs.html.Text";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.graphs.html.VRule";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.graphs.records.RecordField";
                "uqbar.graphs.core.Attachable" -> "uqbar.graphs.html.TableCell";
                "uqbar.graphs.core.Attachable" -> "uqbar.graphs.records.RecordField";
            }
        """
    )
    pickle.dumps(inheritance_graph)


@pytest.mark.xfail(
    sys.version_info.minor >= 11, reason="ReprEnum introduced in 3.11", strict=True
)
def test_03_py310():
    inheritance_graph = uqbar.apis.InheritanceGraph(
        package_paths=["uqbar"], lineage_paths=["uqbar"]
    )
    assert str(inheritance_graph) == uqbar.strings.normalize(
        r"""
            digraph InheritanceGraph {
                graph [bgcolor=transparent,
                    color=lightsteelblue2,
                    fontname=Arial,
                    fontsize=10,
                    outputorder=edgesfirst,
                    overlap=prism,
                    penwidth=2,
                    rankdir=LR,
                    splines=spline,
                    style="dashed, rounded",
                    truecolor=true];
                node [colorscheme=pastel19,
                    fontname=Arial,
                    fontsize=10,
                    height=0,
                    penwidth=2,
                    shape=box,
                    style="filled, rounded",
                    width=0];
                edge [color=lightslategrey,
                    penwidth=1];
                subgraph cluster_builtins {
                    graph [label=builtins];
                    node [color=1];
                    "builtins.BaseException" [label="Base\nException"];
                    "builtins.Exception" [label=Exception];
                    "builtins.int" [label=int];
                    "builtins.object" [label=object];
                    "builtins.BaseException" -> "builtins.Exception";
                    "builtins.object" -> "builtins.BaseException";
                    "builtins.object" -> "builtins.int";
                }
                subgraph cluster_code {
                    graph [label=code];
                    node [color=2];
                    "code.InteractiveConsole" [label="Interactive\nConsole"];
                    "code.InteractiveInterpreter" [label="Interactive\nInterpreter"];
                    "code.InteractiveInterpreter" -> "code.InteractiveConsole";
                }
                subgraph "cluster_collections.abc" {
                    graph [label="collections.abc"];
                    node [color=3];
                    "collections.abc.Collection" [label=Collection,
                        shape=oval,
                        style=bold];
                    "collections.abc.Container" [label=Container,
                        shape=oval,
                        style=bold];
                    "collections.abc.Iterable" [label=Iterable,
                        shape=oval,
                        style=bold];
                    "collections.abc.Mapping" [label=Mapping,
                        shape=oval,
                        style=bold];
                    "collections.abc.MutableMapping" [label="Mutable\nMapping",
                        shape=oval,
                        style=bold];
                    "collections.abc.Sized" [label=Sized,
                        shape=oval,
                        style=bold];
                    "collections.abc.Collection" -> "collections.abc.Mapping";
                    "collections.abc.Container" -> "collections.abc.Collection";
                    "collections.abc.Iterable" -> "collections.abc.Collection";
                    "collections.abc.Mapping" -> "collections.abc.MutableMapping";
                    "collections.abc.Sized" -> "collections.abc.Collection";
                }
                subgraph "cluster_docutils.nodes" {
                    graph [label="docutils.nodes"];
                    node [color=4];
                    "docutils.nodes.Body" [label=Body];
                    "docutils.nodes.Element" [label=Element];
                    "docutils.nodes.General" [label=General];
                    "docutils.nodes.Node" [label="Node"];
                    "docutils.nodes.Body" -> "docutils.nodes.General";
                    "docutils.nodes.Node" -> "docutils.nodes.Element";
                }
                subgraph "cluster_docutils.parsers.rst" {
                    graph [label="docutils.parsers.rst"];
                    node [color=5];
                    "docutils.parsers.rst.Directive" [label=Directive];
                }
                subgraph cluster_enum {
                    graph [label=enum];
                    node [color=6];
                    "enum.Enum" [label=Enum];
                    "enum.IntEnum" [label="Int\nEnum"];
                    "enum.Enum" -> "enum.IntEnum";
                }
                subgraph "cluster_uqbar.apis.builders" {
                    graph [label="uqbar.apis.builders"];
                    node [color=7];
                    "uqbar.apis.builders.APIBuilder" [label=APIBuilder];
                }
                subgraph "cluster_uqbar.apis.documenters" {
                    graph [label="uqbar.apis.documenters"];
                    node [color=8];
                    "uqbar.apis.documenters.ClassDocumenter" [label="Class\nDocumenter"];
                    "uqbar.apis.documenters.FunctionDocumenter" [label="Function\nDocumenter"];
                    "uqbar.apis.documenters.MemberDocumenter" [label="Member\nDocumenter"];
                    "uqbar.apis.documenters.ModuleDocumenter" [label="Module\nDocumenter"];
                    "uqbar.apis.documenters.RootDocumenter" [label="Root\nDocumenter"];
                    "uqbar.apis.documenters.MemberDocumenter" -> "uqbar.apis.documenters.ClassDocumenter";
                    "uqbar.apis.documenters.MemberDocumenter" -> "uqbar.apis.documenters.FunctionDocumenter";
                }
                subgraph "cluster_uqbar.apis.dummy" {
                    graph [label="uqbar.apis.dummy"];
                    node [color=9];
                    "uqbar.apis.dummy.MyChildClass" [label="My\nChild\nClass"];
                    "uqbar.apis.dummy.MyParentClass" [label="My\nParent\nClass"];
                    "uqbar.apis.dummy.MyParentClass" -> "uqbar.apis.dummy.MyChildClass";
                }
                subgraph "cluster_uqbar.apis.graphs" {
                    graph [label="uqbar.apis.graphs"];
                    node [color=1];
                    "uqbar.apis.graphs.InheritanceGraph" [label="Inheritance\nGraph"];
                }
                subgraph "cluster_uqbar.apis.nodes" {
                    graph [label="uqbar.apis.nodes"];
                    node [color=2];
                    "uqbar.apis.nodes.ModuleNode" [label="Module\nNode"];
                    "uqbar.apis.nodes.PackageNode" [label="Package\nNode"];
                }
                subgraph "cluster_uqbar.apis.summarizers" {
                    graph [label="uqbar.apis.summarizers"];
                    node [color=3];
                    "uqbar.apis.summarizers.ImmaterialClassDocumenter" [label="Immaterial\nClass\nDocumenter"];
                    "uqbar.apis.summarizers.ImmaterialModuleDocumenter" [label="Immaterial\nModule\nDocumenter"];
                    "uqbar.apis.summarizers.SummarizingClassDocumenter" [label="Summarizing\nClass\nDocumenter"];
                    "uqbar.apis.summarizers.SummarizingModuleDocumenter" [label="Summarizing\nModule\nDocumenter"];
                    "uqbar.apis.summarizers.SummarizingRootDocumenter" [label="Summarizing\nRoot\nDocumenter"];
                    "uqbar.apis.summarizers.SummarizingClassDocumenter" -> "uqbar.apis.summarizers.ImmaterialClassDocumenter";
                }
                subgraph "cluster_uqbar.book.console" {
                    graph [label="uqbar.book.console"];
                    node [color=4];
                    "uqbar.book.console.Console" [label=Console];
                    "uqbar.book.console.ConsoleError" [label="Console\nError"];
                    "uqbar.book.console.ConsoleInput" [label="Console\nInput"];
                    "uqbar.book.console.ConsoleOutput" [label="Console\nOutput"];
                    "uqbar.book.console.MonkeyPatch" [label="Monkey\nPatch"];
                }
                subgraph "cluster_uqbar.book.extensions" {
                    graph [label="uqbar.book.extensions"];
                    node [color=5];
                    "uqbar.book.extensions.Extension" [label=Extension];
                    "uqbar.book.extensions.GraphExtension" [label="Graph\nExtension"];
                    "uqbar.book.extensions.Extension" -> "uqbar.book.extensions.GraphExtension";
                }
                subgraph "cluster_uqbar.book.sphinx" {
                    graph [label="uqbar.book.sphinx"];
                    node [color=6];
                    "uqbar.book.sphinx.UqbarBookDefaultsDirective" [label="Uqbar\nBook\nDefaults\nDirective"];
                    "uqbar.book.sphinx.UqbarBookDirective" [label="Uqbar\nBook\nDirective"];
                    "uqbar.book.sphinx.UqbarBookImportDirective" [label="Uqbar\nBook\nImport\nDirective"];
                    "uqbar.book.sphinx.uqbar_book_defaults_block" [label="uqbar\nbook\ndefaults\nblock"];
                    "uqbar.book.sphinx.uqbar_book_import_block" [label="uqbar\nbook\nimport\nblock"];
                }
                subgraph "cluster_uqbar.containers.dependency_graph" {
                    graph [label="uqbar.containers.dependency_graph"];
                    node [color=7];
                    "uqbar.containers.dependency_graph.DependencyGraph" [label="Dependency\nGraph"];
                }
                subgraph "cluster_uqbar.containers.unique_tree" {
                    graph [label="uqbar.containers.unique_tree"];
                    node [color=8];
                    "uqbar.containers.unique_tree.UniqueTreeContainer" [label="Unique\nTree\nContainer"];
                    "uqbar.containers.unique_tree.UniqueTreeDict" [label="Unique\nTree\nDict"];
                    "uqbar.containers.unique_tree.UniqueTreeList" [label="Unique\nTree\nList"];
                    "uqbar.containers.unique_tree.UniqueTreeNode" [label="Unique\nTree\nNode"];
                    "uqbar.containers.unique_tree.UniqueTreeSet" [label="Unique\nTree\nSet"];
                    "uqbar.containers.unique_tree.UniqueTreeTuple" [label="Unique\nTree\nTuple"];
                    "uqbar.containers.unique_tree.UniqueTreeContainer" -> "uqbar.containers.unique_tree.UniqueTreeDict";
                    "uqbar.containers.unique_tree.UniqueTreeContainer" -> "uqbar.containers.unique_tree.UniqueTreeList";
                    "uqbar.containers.unique_tree.UniqueTreeContainer" -> "uqbar.containers.unique_tree.UniqueTreeSet";
                    "uqbar.containers.unique_tree.UniqueTreeContainer" -> "uqbar.containers.unique_tree.UniqueTreeTuple";
                    "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.containers.unique_tree.UniqueTreeContainer";
                }
                subgraph "cluster_uqbar.enums" {
                    graph [label="uqbar.enums"];
                    node [color=9];
                    "uqbar.enums.IntEnumeration" [label="Int\nEnumeration"];
                    "uqbar.enums.StrictEnumeration" [label="Strict\nEnumeration"];
                }
                subgraph "cluster_uqbar.graphs.attrs" {
                    graph [label="uqbar.graphs.attrs"];
                    node [color=1];
                    "uqbar.graphs.attrs.Attributes" [label=Attributes];
                }
                subgraph "cluster_uqbar.graphs.core" {
                    graph [label="uqbar.graphs.core"];
                    node [color=2];
                    "uqbar.graphs.core.Attachable" [label=Attachable];
                    "uqbar.graphs.core.Edge" [label="Edge"];
                    "uqbar.graphs.core.Graph" [label="Graph"];
                    "uqbar.graphs.core.Node" [label="Node"];
                }
                subgraph "cluster_uqbar.graphs.graphers" {
                    graph [label="uqbar.graphs.graphers"];
                    node [color=3];
                    "uqbar.graphs.graphers.Grapher" [label=Grapher];
                }
                subgraph "cluster_uqbar.graphs.html" {
                    graph [label="uqbar.graphs.html"];
                    node [color=4];
                    "uqbar.graphs.html.HRule" [label=HRule];
                    "uqbar.graphs.html.LineBreak" [label="Line\nBreak"];
                    "uqbar.graphs.html.Table" [label=Table];
                    "uqbar.graphs.html.TableCell" [label="Table\nCell"];
                    "uqbar.graphs.html.TableRow" [label="Table\nRow"];
                    "uqbar.graphs.html.Text" [label=Text];
                    "uqbar.graphs.html.VRule" [label=VRule];
                }
                subgraph "cluster_uqbar.graphs.records" {
                    graph [label="uqbar.graphs.records"];
                    node [color=5];
                    "uqbar.graphs.records.RecordField" [label="Record\nField"];
                    "uqbar.graphs.records.RecordGroup" [label="Record\nGroup"];
                }
                subgraph "cluster_uqbar.io" {
                    graph [label="uqbar.io"];
                    node [color=6];
                    "uqbar.io.DirectoryChange" [color=black,
                        fontcolor=white,
                        label="Directory\nChange"];
                    "uqbar.io.Profiler" [color=black,
                        fontcolor=white,
                        label=Profiler];
                    "uqbar.io.RedirectedStreams" [color=black,
                        fontcolor=white,
                        label="Redirected\nStreams"];
                    "uqbar.io.Timer" [color=black,
                        fontcolor=white,
                        label=Timer];
                }
                subgraph "cluster_uqbar.sphinx.inheritance" {
                    graph [label="uqbar.sphinx.inheritance"];
                    node [color=7];
                    "uqbar.sphinx.inheritance.InheritanceDiagram" [label="Inheritance\nDiagram"];
                    "uqbar.sphinx.inheritance.inheritance_diagram" [label="inheritance\ndiagram"];
                }
                "builtins.Exception" -> "uqbar.book.console.ConsoleError";
                "builtins.int" -> "enum.IntEnum";
                "builtins.object" -> "code.InteractiveInterpreter";
                "builtins.object" -> "collections.abc.Container";
                "builtins.object" -> "collections.abc.Iterable";
                "builtins.object" -> "collections.abc.Sized";
                "builtins.object" -> "docutils.nodes.Body";
                "builtins.object" -> "docutils.nodes.Node";
                "builtins.object" -> "docutils.parsers.rst.Directive";
                "builtins.object" -> "enum.Enum";
                "builtins.object" -> "uqbar.apis.builders.APIBuilder";
                "builtins.object" -> "uqbar.apis.documenters.MemberDocumenter";
                "builtins.object" -> "uqbar.apis.documenters.ModuleDocumenter";
                "builtins.object" -> "uqbar.apis.documenters.RootDocumenter";
                "builtins.object" -> "uqbar.apis.dummy.MyParentClass";
                "builtins.object" -> "uqbar.apis.graphs.InheritanceGraph";
                "builtins.object" -> "uqbar.book.console.ConsoleInput";
                "builtins.object" -> "uqbar.book.console.ConsoleOutput";
                "builtins.object" -> "uqbar.book.console.MonkeyPatch";
                "builtins.object" -> "uqbar.book.extensions.Extension";
                "builtins.object" -> "uqbar.containers.dependency_graph.DependencyGraph";
                "builtins.object" -> "uqbar.containers.unique_tree.UniqueTreeNode";
                "builtins.object" -> "uqbar.graphs.core.Edge";
                "builtins.object" -> "uqbar.graphs.graphers.Grapher";
                "builtins.object" -> "uqbar.io.DirectoryChange";
                "builtins.object" -> "uqbar.io.Profiler";
                "builtins.object" -> "uqbar.io.RedirectedStreams";
                "builtins.object" -> "uqbar.io.Timer";
                "code.InteractiveConsole" -> "uqbar.book.console.Console";
                "collections.abc.MutableMapping" -> "uqbar.graphs.attrs.Attributes";
                "docutils.nodes.Element" -> "uqbar.book.sphinx.uqbar_book_defaults_block";
                "docutils.nodes.Element" -> "uqbar.book.sphinx.uqbar_book_import_block";
                "docutils.nodes.Element" -> "uqbar.sphinx.inheritance.inheritance_diagram";
                "docutils.nodes.General" -> "uqbar.book.sphinx.uqbar_book_defaults_block";
                "docutils.nodes.General" -> "uqbar.book.sphinx.uqbar_book_import_block";
                "docutils.nodes.General" -> "uqbar.sphinx.inheritance.inheritance_diagram";
                "docutils.parsers.rst.Directive" -> "uqbar.book.sphinx.UqbarBookDefaultsDirective";
                "docutils.parsers.rst.Directive" -> "uqbar.book.sphinx.UqbarBookDirective";
                "docutils.parsers.rst.Directive" -> "uqbar.book.sphinx.UqbarBookImportDirective";
                "docutils.parsers.rst.Directive" -> "uqbar.sphinx.inheritance.InheritanceDiagram";
                "enum.Enum" -> "uqbar.enums.StrictEnumeration";
                "enum.IntEnum" -> "uqbar.enums.IntEnumeration";
                "uqbar.apis.documenters.ClassDocumenter" -> "uqbar.apis.summarizers.SummarizingClassDocumenter";
                "uqbar.apis.documenters.ModuleDocumenter" -> "uqbar.apis.summarizers.ImmaterialModuleDocumenter";
                "uqbar.apis.documenters.ModuleDocumenter" -> "uqbar.apis.summarizers.SummarizingModuleDocumenter";
                "uqbar.apis.documenters.RootDocumenter" -> "uqbar.apis.summarizers.SummarizingRootDocumenter";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.apis.nodes.PackageNode";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.graphs.core.Graph";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.graphs.core.Node";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.graphs.html.Table";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.graphs.html.TableCell";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.graphs.html.TableRow";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.graphs.records.RecordGroup";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.apis.nodes.ModuleNode";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.graphs.core.Attachable";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.graphs.html.HRule";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.graphs.html.LineBreak";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.graphs.html.Text";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.graphs.html.VRule";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.graphs.records.RecordField";
                "uqbar.graphs.core.Attachable" -> "uqbar.graphs.html.TableCell";
                "uqbar.graphs.core.Attachable" -> "uqbar.graphs.records.RecordField";
            }
        """
    )
    pickle.dumps(inheritance_graph)


@pytest.mark.xfail(
    sys.version_info.minor < 11, reason="ReprEnum introduced in 3.11", strict=True
)
def test_03_py311():
    inheritance_graph = uqbar.apis.InheritanceGraph(
        package_paths=["uqbar"], lineage_paths=["uqbar"]
    )
    assert str(inheritance_graph) == uqbar.strings.normalize(
        r"""
            digraph InheritanceGraph {
                graph [bgcolor=transparent,
                    color=lightsteelblue2,
                    fontname=Arial,
                    fontsize=10,
                    outputorder=edgesfirst,
                    overlap=prism,
                    penwidth=2,
                    rankdir=LR,
                    splines=spline,
                    style="dashed, rounded",
                    truecolor=true];
                node [colorscheme=pastel19,
                    fontname=Arial,
                    fontsize=10,
                    height=0,
                    penwidth=2,
                    shape=box,
                    style="filled, rounded",
                    width=0];
                edge [color=lightslategrey,
                    penwidth=1];
                subgraph cluster_builtins {
                    graph [label=builtins];
                    node [color=1];
                    "builtins.BaseException" [label="Base\nException"];
                    "builtins.Exception" [label=Exception];
                    "builtins.int" [label=int];
                    "builtins.object" [label=object];
                    "builtins.BaseException" -> "builtins.Exception";
                    "builtins.object" -> "builtins.BaseException";
                    "builtins.object" -> "builtins.int";
                }
                subgraph cluster_code {
                    graph [label=code];
                    node [color=2];
                    "code.InteractiveConsole" [label="Interactive\nConsole"];
                    "code.InteractiveInterpreter" [label="Interactive\nInterpreter"];
                    "code.InteractiveInterpreter" -> "code.InteractiveConsole";
                }
                subgraph "cluster_collections.abc" {
                    graph [label="collections.abc"];
                    node [color=3];
                    "collections.abc.Collection" [label=Collection,
                        shape=oval,
                        style=bold];
                    "collections.abc.Container" [label=Container,
                        shape=oval,
                        style=bold];
                    "collections.abc.Iterable" [label=Iterable,
                        shape=oval,
                        style=bold];
                    "collections.abc.Mapping" [label=Mapping,
                        shape=oval,
                        style=bold];
                    "collections.abc.MutableMapping" [label="Mutable\nMapping",
                        shape=oval,
                        style=bold];
                    "collections.abc.Sized" [label=Sized,
                        shape=oval,
                        style=bold];
                    "collections.abc.Collection" -> "collections.abc.Mapping";
                    "collections.abc.Container" -> "collections.abc.Collection";
                    "collections.abc.Iterable" -> "collections.abc.Collection";
                    "collections.abc.Mapping" -> "collections.abc.MutableMapping";
                    "collections.abc.Sized" -> "collections.abc.Collection";
                }
                subgraph "cluster_docutils.nodes" {
                    graph [label="docutils.nodes"];
                    node [color=4];
                    "docutils.nodes.Body" [label=Body];
                    "docutils.nodes.Element" [label=Element];
                    "docutils.nodes.General" [label=General];
                    "docutils.nodes.Node" [label="Node"];
                    "docutils.nodes.Body" -> "docutils.nodes.General";
                    "docutils.nodes.Node" -> "docutils.nodes.Element";
                }
                subgraph "cluster_docutils.parsers.rst" {
                    graph [label="docutils.parsers.rst"];
                    node [color=5];
                    "docutils.parsers.rst.Directive" [label=Directive];
                }
                subgraph cluster_enum {
                    graph [label=enum];
                    node [color=6];
                    "enum.Enum" [label=Enum];
                    "enum.IntEnum" [label="Int\nEnum"];
                    "enum.ReprEnum" [label="Repr\nEnum"];
                    "enum.Enum" -> "enum.ReprEnum";
                    "enum.ReprEnum" -> "enum.IntEnum";
                }
                subgraph "cluster_uqbar.apis.builders" {
                    graph [label="uqbar.apis.builders"];
                    node [color=7];
                    "uqbar.apis.builders.APIBuilder" [label=APIBuilder];
                }
                subgraph "cluster_uqbar.apis.documenters" {
                    graph [label="uqbar.apis.documenters"];
                    node [color=8];
                    "uqbar.apis.documenters.ClassDocumenter" [label="Class\nDocumenter"];
                    "uqbar.apis.documenters.FunctionDocumenter" [label="Function\nDocumenter"];
                    "uqbar.apis.documenters.MemberDocumenter" [label="Member\nDocumenter"];
                    "uqbar.apis.documenters.ModuleDocumenter" [label="Module\nDocumenter"];
                    "uqbar.apis.documenters.RootDocumenter" [label="Root\nDocumenter"];
                    "uqbar.apis.documenters.MemberDocumenter" -> "uqbar.apis.documenters.ClassDocumenter";
                    "uqbar.apis.documenters.MemberDocumenter" -> "uqbar.apis.documenters.FunctionDocumenter";
                }
                subgraph "cluster_uqbar.apis.dummy" {
                    graph [label="uqbar.apis.dummy"];
                    node [color=9];
                    "uqbar.apis.dummy.MyChildClass" [label="My\nChild\nClass"];
                    "uqbar.apis.dummy.MyParentClass" [label="My\nParent\nClass"];
                    "uqbar.apis.dummy.MyParentClass" -> "uqbar.apis.dummy.MyChildClass";
                }
                subgraph "cluster_uqbar.apis.graphs" {
                    graph [label="uqbar.apis.graphs"];
                    node [color=1];
                    "uqbar.apis.graphs.InheritanceGraph" [label="Inheritance\nGraph"];
                }
                subgraph "cluster_uqbar.apis.nodes" {
                    graph [label="uqbar.apis.nodes"];
                    node [color=2];
                    "uqbar.apis.nodes.ModuleNode" [label="Module\nNode"];
                    "uqbar.apis.nodes.PackageNode" [label="Package\nNode"];
                }
                subgraph "cluster_uqbar.apis.summarizers" {
                    graph [label="uqbar.apis.summarizers"];
                    node [color=3];
                    "uqbar.apis.summarizers.ImmaterialClassDocumenter" [label="Immaterial\nClass\nDocumenter"];
                    "uqbar.apis.summarizers.ImmaterialModuleDocumenter" [label="Immaterial\nModule\nDocumenter"];
                    "uqbar.apis.summarizers.SummarizingClassDocumenter" [label="Summarizing\nClass\nDocumenter"];
                    "uqbar.apis.summarizers.SummarizingModuleDocumenter" [label="Summarizing\nModule\nDocumenter"];
                    "uqbar.apis.summarizers.SummarizingRootDocumenter" [label="Summarizing\nRoot\nDocumenter"];
                    "uqbar.apis.summarizers.SummarizingClassDocumenter" -> "uqbar.apis.summarizers.ImmaterialClassDocumenter";
                }
                subgraph "cluster_uqbar.book.console" {
                    graph [label="uqbar.book.console"];
                    node [color=4];
                    "uqbar.book.console.Console" [label=Console];
                    "uqbar.book.console.ConsoleError" [label="Console\nError"];
                    "uqbar.book.console.ConsoleInput" [label="Console\nInput"];
                    "uqbar.book.console.ConsoleOutput" [label="Console\nOutput"];
                    "uqbar.book.console.MonkeyPatch" [label="Monkey\nPatch"];
                }
                subgraph "cluster_uqbar.book.extensions" {
                    graph [label="uqbar.book.extensions"];
                    node [color=5];
                    "uqbar.book.extensions.Extension" [label=Extension];
                    "uqbar.book.extensions.GraphExtension" [label="Graph\nExtension"];
                    "uqbar.book.extensions.Extension" -> "uqbar.book.extensions.GraphExtension";
                }
                subgraph "cluster_uqbar.book.sphinx" {
                    graph [label="uqbar.book.sphinx"];
                    node [color=6];
                    "uqbar.book.sphinx.UqbarBookDefaultsDirective" [label="Uqbar\nBook\nDefaults\nDirective"];
                    "uqbar.book.sphinx.UqbarBookDirective" [label="Uqbar\nBook\nDirective"];
                    "uqbar.book.sphinx.UqbarBookImportDirective" [label="Uqbar\nBook\nImport\nDirective"];
                    "uqbar.book.sphinx.uqbar_book_defaults_block" [label="uqbar\nbook\ndefaults\nblock"];
                    "uqbar.book.sphinx.uqbar_book_import_block" [label="uqbar\nbook\nimport\nblock"];
                }
                subgraph "cluster_uqbar.containers.dependency_graph" {
                    graph [label="uqbar.containers.dependency_graph"];
                    node [color=7];
                    "uqbar.containers.dependency_graph.DependencyGraph" [label="Dependency\nGraph"];
                }
                subgraph "cluster_uqbar.containers.unique_tree" {
                    graph [label="uqbar.containers.unique_tree"];
                    node [color=8];
                    "uqbar.containers.unique_tree.UniqueTreeContainer" [label="Unique\nTree\nContainer"];
                    "uqbar.containers.unique_tree.UniqueTreeDict" [label="Unique\nTree\nDict"];
                    "uqbar.containers.unique_tree.UniqueTreeList" [label="Unique\nTree\nList"];
                    "uqbar.containers.unique_tree.UniqueTreeNode" [label="Unique\nTree\nNode"];
                    "uqbar.containers.unique_tree.UniqueTreeSet" [label="Unique\nTree\nSet"];
                    "uqbar.containers.unique_tree.UniqueTreeTuple" [label="Unique\nTree\nTuple"];
                    "uqbar.containers.unique_tree.UniqueTreeContainer" -> "uqbar.containers.unique_tree.UniqueTreeDict";
                    "uqbar.containers.unique_tree.UniqueTreeContainer" -> "uqbar.containers.unique_tree.UniqueTreeList";
                    "uqbar.containers.unique_tree.UniqueTreeContainer" -> "uqbar.containers.unique_tree.UniqueTreeSet";
                    "uqbar.containers.unique_tree.UniqueTreeContainer" -> "uqbar.containers.unique_tree.UniqueTreeTuple";
                    "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.containers.unique_tree.UniqueTreeContainer";
                }
                subgraph "cluster_uqbar.enums" {
                    graph [label="uqbar.enums"];
                    node [color=9];
                    "uqbar.enums.IntEnumeration" [label="Int\nEnumeration"];
                    "uqbar.enums.StrictEnumeration" [label="Strict\nEnumeration"];
                }
                subgraph "cluster_uqbar.graphs.attrs" {
                    graph [label="uqbar.graphs.attrs"];
                    node [color=1];
                    "uqbar.graphs.attrs.Attributes" [label=Attributes];
                }
                subgraph "cluster_uqbar.graphs.core" {
                    graph [label="uqbar.graphs.core"];
                    node [color=2];
                    "uqbar.graphs.core.Attachable" [label=Attachable];
                    "uqbar.graphs.core.Edge" [label="Edge"];
                    "uqbar.graphs.core.Graph" [label="Graph"];
                    "uqbar.graphs.core.Node" [label="Node"];
                }
                subgraph "cluster_uqbar.graphs.graphers" {
                    graph [label="uqbar.graphs.graphers"];
                    node [color=3];
                    "uqbar.graphs.graphers.Grapher" [label=Grapher];
                }
                subgraph "cluster_uqbar.graphs.html" {
                    graph [label="uqbar.graphs.html"];
                    node [color=4];
                    "uqbar.graphs.html.HRule" [label=HRule];
                    "uqbar.graphs.html.LineBreak" [label="Line\nBreak"];
                    "uqbar.graphs.html.Table" [label=Table];
                    "uqbar.graphs.html.TableCell" [label="Table\nCell"];
                    "uqbar.graphs.html.TableRow" [label="Table\nRow"];
                    "uqbar.graphs.html.Text" [label=Text];
                    "uqbar.graphs.html.VRule" [label=VRule];
                }
                subgraph "cluster_uqbar.graphs.records" {
                    graph [label="uqbar.graphs.records"];
                    node [color=5];
                    "uqbar.graphs.records.RecordField" [label="Record\nField"];
                    "uqbar.graphs.records.RecordGroup" [label="Record\nGroup"];
                }
                subgraph "cluster_uqbar.io" {
                    graph [label="uqbar.io"];
                    node [color=6];
                    "uqbar.io.DirectoryChange" [color=black,
                        fontcolor=white,
                        label="Directory\nChange"];
                    "uqbar.io.Profiler" [color=black,
                        fontcolor=white,
                        label=Profiler];
                    "uqbar.io.RedirectedStreams" [color=black,
                        fontcolor=white,
                        label="Redirected\nStreams"];
                    "uqbar.io.Timer" [color=black,
                        fontcolor=white,
                        label=Timer];
                }
                subgraph "cluster_uqbar.sphinx.inheritance" {
                    graph [label="uqbar.sphinx.inheritance"];
                    node [color=7];
                    "uqbar.sphinx.inheritance.InheritanceDiagram" [label="Inheritance\nDiagram"];
                    "uqbar.sphinx.inheritance.inheritance_diagram" [label="inheritance\ndiagram"];
                }
                "builtins.Exception" -> "uqbar.book.console.ConsoleError";
                "builtins.int" -> "enum.IntEnum";
                "builtins.object" -> "code.InteractiveInterpreter";
                "builtins.object" -> "collections.abc.Container";
                "builtins.object" -> "collections.abc.Iterable";
                "builtins.object" -> "collections.abc.Sized";
                "builtins.object" -> "docutils.nodes.Body";
                "builtins.object" -> "docutils.nodes.Node";
                "builtins.object" -> "docutils.parsers.rst.Directive";
                "builtins.object" -> "enum.Enum";
                "builtins.object" -> "uqbar.apis.builders.APIBuilder";
                "builtins.object" -> "uqbar.apis.documenters.MemberDocumenter";
                "builtins.object" -> "uqbar.apis.documenters.ModuleDocumenter";
                "builtins.object" -> "uqbar.apis.documenters.RootDocumenter";
                "builtins.object" -> "uqbar.apis.dummy.MyParentClass";
                "builtins.object" -> "uqbar.apis.graphs.InheritanceGraph";
                "builtins.object" -> "uqbar.book.console.ConsoleInput";
                "builtins.object" -> "uqbar.book.console.ConsoleOutput";
                "builtins.object" -> "uqbar.book.console.MonkeyPatch";
                "builtins.object" -> "uqbar.book.extensions.Extension";
                "builtins.object" -> "uqbar.containers.dependency_graph.DependencyGraph";
                "builtins.object" -> "uqbar.containers.unique_tree.UniqueTreeNode";
                "builtins.object" -> "uqbar.graphs.core.Edge";
                "builtins.object" -> "uqbar.graphs.graphers.Grapher";
                "builtins.object" -> "uqbar.io.DirectoryChange";
                "builtins.object" -> "uqbar.io.Profiler";
                "builtins.object" -> "uqbar.io.RedirectedStreams";
                "builtins.object" -> "uqbar.io.Timer";
                "code.InteractiveConsole" -> "uqbar.book.console.Console";
                "collections.abc.MutableMapping" -> "uqbar.graphs.attrs.Attributes";
                "docutils.nodes.Element" -> "uqbar.book.sphinx.uqbar_book_defaults_block";
                "docutils.nodes.Element" -> "uqbar.book.sphinx.uqbar_book_import_block";
                "docutils.nodes.Element" -> "uqbar.sphinx.inheritance.inheritance_diagram";
                "docutils.nodes.General" -> "uqbar.book.sphinx.uqbar_book_defaults_block";
                "docutils.nodes.General" -> "uqbar.book.sphinx.uqbar_book_import_block";
                "docutils.nodes.General" -> "uqbar.sphinx.inheritance.inheritance_diagram";
                "docutils.parsers.rst.Directive" -> "uqbar.book.sphinx.UqbarBookDefaultsDirective";
                "docutils.parsers.rst.Directive" -> "uqbar.book.sphinx.UqbarBookDirective";
                "docutils.parsers.rst.Directive" -> "uqbar.book.sphinx.UqbarBookImportDirective";
                "docutils.parsers.rst.Directive" -> "uqbar.sphinx.inheritance.InheritanceDiagram";
                "enum.Enum" -> "uqbar.enums.StrictEnumeration";
                "enum.IntEnum" -> "uqbar.enums.IntEnumeration";
                "uqbar.apis.documenters.ClassDocumenter" -> "uqbar.apis.summarizers.SummarizingClassDocumenter";
                "uqbar.apis.documenters.ModuleDocumenter" -> "uqbar.apis.summarizers.ImmaterialModuleDocumenter";
                "uqbar.apis.documenters.ModuleDocumenter" -> "uqbar.apis.summarizers.SummarizingModuleDocumenter";
                "uqbar.apis.documenters.RootDocumenter" -> "uqbar.apis.summarizers.SummarizingRootDocumenter";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.apis.nodes.PackageNode";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.graphs.core.Graph";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.graphs.core.Node";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.graphs.html.Table";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.graphs.html.TableCell";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.graphs.html.TableRow";
                "uqbar.containers.unique_tree.UniqueTreeList" -> "uqbar.graphs.records.RecordGroup";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.apis.nodes.ModuleNode";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.graphs.core.Attachable";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.graphs.html.HRule";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.graphs.html.LineBreak";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.graphs.html.Text";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.graphs.html.VRule";
                "uqbar.containers.unique_tree.UniqueTreeNode" -> "uqbar.graphs.records.RecordField";
                "uqbar.graphs.core.Attachable" -> "uqbar.graphs.html.TableCell";
                "uqbar.graphs.core.Attachable" -> "uqbar.graphs.records.RecordField";
            }
        """
    )
    pickle.dumps(inheritance_graph)


def test_04():
    inheritance_graph = uqbar.apis.InheritanceGraph(
        package_paths=["uqbar"], lineage_paths=["uqbar.apis.dummy"]
    )
    assert inheritance_graph.parents_to_children == collections.OrderedDict(
        [
            ("builtins.object", ["uqbar.apis.dummy.MyParentClass"]),
            ("uqbar.apis.dummy.MyParentClass", ["uqbar.apis.dummy.MyChildClass"]),
        ]
    )
    assert str(inheritance_graph) == uqbar.strings.normalize(
        r"""
        digraph InheritanceGraph {
            graph [bgcolor=transparent,
                color=lightsteelblue2,
                fontname=Arial,
                fontsize=10,
                outputorder=edgesfirst,
                overlap=prism,
                penwidth=2,
                rankdir=LR,
                splines=spline,
                style="dashed, rounded",
                truecolor=true];
            node [colorscheme=pastel19,
                fontname=Arial,
                fontsize=10,
                height=0,
                penwidth=2,
                shape=box,
                style="filled, rounded",
                width=0];
            edge [color=lightslategrey,
                penwidth=1];
            subgraph cluster_builtins {
                graph [label=builtins];
                node [color=1];
                "builtins.object" [label=object];
            }
            subgraph "cluster_uqbar.apis.dummy" {
                graph [label="uqbar.apis.dummy"];
                node [color=2];
                "uqbar.apis.dummy.MyChildClass" [color=black,
                    fontcolor=white,
                    label="My\nChild\nClass"];
                "uqbar.apis.dummy.MyParentClass" [color=black,
                    fontcolor=white,
                    label="My\nParent\nClass"];
                "uqbar.apis.dummy.MyParentClass" -> "uqbar.apis.dummy.MyChildClass";
            }
            "builtins.object" -> "uqbar.apis.dummy.MyParentClass";
        }
    """
    )
    pickle.dumps(inheritance_graph)
    pickle.dumps(uqbar.apis.dummy.MyChildClass)
    with pytest.raises(AttributeError):
        pickle.dumps(uqbar.apis.dummy.MyParentClass)



import shutil
import sys
from pathlib import Path

import pytest

pytest_plugins = "sphinx.testing.fixtures"


collect_ignore = ["roots"]


@pytest.fixture(scope="session")
def remove_sphinx_projects(sphinx_test_tempdir):
    # Even upon exception, remove any directory from temp area
    # which looks like a Sphinx project. This ONLY runs once.
    roots_path = Path(sphinx_test_tempdir)
    for path in roots_path.iterdir():
        if path.is_dir():
            if (Path(path) / "_build").exists():
                # This directory is a Sphinx project, remove it
                shutil.rmtree(path)


@pytest.fixture()
def rootdir(remove_sphinx_projects):
    root_path = Path(__file__).parent / "roots"
    if sys.version_info > (3, 8):
        from sphinx.testing.path import path

        yield path(str(root_path))
    else:
        yield root_path



from unittest import mock

from uqbar import graphs


def test_01(tmp_path):
    graph = graphs.Graph()
    node_a = graphs.Node()
    node_b = graphs.Node()
    node_a.attach(node_b)
    graph.extend([node_a, node_b])
    assert list(tmp_path.iterdir()) == []
    with mock.patch.object(graphs.Grapher, "open_output_path") as patcher:
        graphs.Grapher(graph, output_directory=tmp_path)()
    assert patcher.call_count == 1
    for glob in ["*.dot", "*.log", "*.pdf"]:
        assert len(list(tmp_path.glob(glob))) == 1



import collections
import contextlib
import importlib
import inspect
import itertools
import logging
import pickle
import sqlite3
import traceback
from typing import Any, Dict

from docutils.frontend import get_default_settings
from docutils.nodes import Element, General, doctest_block, literal_block
from docutils.parsers.rst import Directive, Parser
from docutils.parsers.rst.directives import flag, register_directive
from docutils.utils import new_document
from sphinx.util.nodes import set_source_info

from .console import Console, ConsoleError, ConsoleInput, ConsoleOutput

logger = logging.getLogger(__name__)

try:
    import black

    def black_format(lines):
        mode = black.FileMode(
            line_length=80, target_versions=set([black.TargetVersion.PY36])
        )
        return black.format_str("\n".join(lines), mode=mode).splitlines()

except ImportError:

    def black_format(lines):
        return lines


class UqbarBookDirective(Directive):
    __documentation_ignore_inherited__ = True

    has_content = True
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec: Dict[str, Any] = {"allow-exceptions": flag, "hide": flag}

    def run(self):
        self.assert_has_content()
        code = "\n".join(self.content)
        block = literal_block(code, code)
        block.line = self.content_offset  # set the content line number
        set_source_info(self, block)
        for key, spec in self.option_spec.items():
            if key not in self.options:
                continue
            option = self.options[key]
            if spec == flag:
                option = True
            block[key] = option
        return [block]


class UqbarBookDefaultsDirective(Directive):
    __documentation_ignore_inherited__ = True

    has_content = False
    required_arguments = 0
    optional_arguments = 0
    option_spec: Dict[str, Any] = {}

    def run(self):
        block = uqbar_book_defaults_block()
        for key, value in self.options.items():
            block[key] = value
        return [block]


class UqbarBookImportDirective(Directive):
    __documentation_ignore_inherited__ = True

    has_content = False
    required_arguments = 1
    optional_arguments = 0
    option_spec = {"hide": flag}

    def run(self):
        block = uqbar_book_import_block()
        block["path"] = self.arguments[0]
        block.line = self.content_offset  # set the content line number
        set_source_info(self, block)
        if "hide" in self.options:
            block["hide"] = True
        return [block]


class uqbar_book_defaults_block(General, Element):
    __documentation_ignore_inherited__ = True


class uqbar_book_import_block(General, Element):
    __documentation_ignore_inherited__ = True


def collect_literal_blocks(document):
    prototype = (
        literal_block,
        doctest_block,
        uqbar_book_defaults_block,
        uqbar_book_import_block,
    )
    blocks = []
    for block in document.findall(lambda node: isinstance(node, prototype)):
        if not isinstance(block, (uqbar_book_defaults_block, uqbar_book_import_block)):
            contents = block[0]
            if isinstance(contents, literal_block):
                contents = contents[0]
            if not contents.strip().startswith(">>>"):
                continue
        blocks.append(block)
    return blocks


def create_cache_db(path):
    connection = sqlite3.connect(str(path))
    connection.execute("CREATE TABLE cache (path TEXT, data BLOB, hits NUMBER);")
    connection.commit()
    return connection


def query_cache_db(connection, cache_path):
    result = None
    for row in connection.execute(
        "SELECT data, hits FROM cache WHERE path = ?", (cache_path,)
    ):
        result, hits = row
    if not result:
        return None
    connection.execute(
        "UPDATE cache SET hits = ? WHERE path = ?", (hits + 1, cache_path)
    )
    connection.commit()
    return pickle.loads(result)


def update_cache_db(connection, cache_path, data):
    connection.execute(
        "INSERT INTO cache VALUES (?, ?, 0)", (cache_path, pickle.dumps(data))
    )
    connection.commit()


def group_literal_blocks_by_cache_path(blocks):
    cache_mapping = {None: [[]]}
    last_cache_path = None
    for block in blocks:
        cache_path = literal_block_to_cache_path(block)
        if cache_path is None:
            cache_mapping[None][0].append(block)
            last_cache_path = None
        else:
            if cache_path != last_cache_path:
                block_list = []
                cache_mapping.setdefault(cache_path, []).append(block_list)
            cache_mapping[cache_path][-1].append(block)
            last_cache_path = cache_path
    return cache_mapping


def find_traceback(console_output):
    for item in reversed(console_output):
        if not isinstance(item, ConsoleOutput):
            continue
        # There could be non-traceback output before the traceback appears.
        lines = item.string.splitlines()
        while lines:
            if lines[0].startswith("Traceback (most recent call last):"):
                break
            lines.pop(0)
        if lines:
            return "\n".join(lines) + "\n"


@contextlib.contextmanager
def console_context(
    *, extensions, namespace=None, setup_lines=None, teardown_lines=None, document=None
):
    console = Console(extensions=extensions, namespace=namespace)
    console_output, errored = console.interpret(setup_lines or [])
    if errored:
        raise ConsoleError(find_traceback(console_output), document)
    try:
        yield console
    finally:
        console.resetbuffer()
        console_output, errored = console.interpret(teardown_lines or [])
        if errored:
            raise ConsoleError(find_traceback(console_output), document)


def interpret_code_blocks(
    blocks,
    allow_exceptions=False,
    document=None,
    extensions=None,
    namespace=None,
    setup_lines=None,
    teardown_lines=None,
    logger_func=None,
    use_black=False,
):
    results = collections.OrderedDict()
    block = blocks[0] if blocks else None
    with console_context(
        document=document,
        extensions=extensions,
        namespace=namespace,
        setup_lines=setup_lines,
        teardown_lines=teardown_lines,
    ) as console:
        default_proxy_options = {}
        for block in blocks:
            if isinstance(block, uqbar_book_defaults_block):
                default_proxy_options = dict(block.attlist())
                continue
            else:
                proxy_options = {**default_proxy_options, **dict(block.attlist())}
                console.push_proxy_options(proxy_options)
            if isinstance(block, uqbar_book_import_block):
                console_output, errored, has_exception = interpret_import_block(
                    console, block
                )
            elif isinstance(block, (literal_block, doctest_block)):
                console_output, errored, has_exception = interpret_literal_block(
                    console, block, use_black=use_black
                )
            if errored:
                traceback = find_traceback(console_output)
                if logger_func:
                    logger_func(traceback)
                if not (
                    block.get("allow-exceptions") or allow_exceptions or has_exception
                ):
                    raise ConsoleError(traceback, block)
            if block.get("hide"):
                console_output = [
                    _
                    for _ in console_output
                    if not isinstance(_, (ConsoleInput, ConsoleOutput))
                ]
            results[block] = console_output
    return results


def interpret_code_blocks_with_cache(
    blocks,
    cache_path,
    connection,
    allow_exceptions=False,
    document=None,
    extensions=None,
    namespace=None,
    setup_lines=None,
    teardown_lines=None,
    use_black=False,
):
    cache_data = query_cache_db(connection, cache_path)
    if cache_data is not None:
        local_node_mapping = collections.OrderedDict(zip(blocks, cache_data))
    else:
        local_node_mapping = interpret_code_blocks(
            blocks,
            allow_exceptions=allow_exceptions,
            document=document,
            extensions=extensions,
            namespace=namespace,
            setup_lines=setup_lines,
            teardown_lines=teardown_lines,
            use_black=use_black,
        )
        if cache_path is not None:
            cache_data = list(local_node_mapping.values())
            update_cache_db(connection, cache_path, cache_data)
    return local_node_mapping


def interpret_import_block(console, block):
    code_address = block["path"]
    module_name, sep, attr_name = code_address.rpartition(":")
    module = importlib.import_module(module_name)
    attr = getattr(module, attr_name)
    source = inspect.getsource(attr)
    lines = [f"from {module_name} import {attr_name}"]
    _, errored = console.interpret(lines)
    output = [ConsoleOutput(string=source)]
    return output, errored, False


def interpret_literal_block(console, block, use_black=False):
    has_exception = False
    lines = []
    contents = block[0]
    if isinstance(contents, literal_block):
        contents = contents[0]
    for line in contents.splitlines():
        if line.startswith((">>> ", "... ")):
            lines.append(line[4:])
        elif line.rstrip() == "...":
            lines.append("")
        elif line.startswith("Traceback (most recent call last):"):
            has_exception = True
    if use_black:
        try:
            lines = black_format(lines)
        except Exception:
            raise ConsoleError(traceback.format_exc(), block)
    console_output, errored = console.interpret(lines)
    return console_output, errored, has_exception


def literal_block_to_cache_path(block):
    desc = None
    parent = block.parent
    while desc is None and parent is not None:
        if parent.tagname == "desc":
            desc = parent
            break
        else:
            parent = parent.parent
    if desc is None:
        return None
    desc_signature = desc[0]
    module_path = desc_signature.attributes["module"]
    object_path = desc_signature.attributes["fullname"]
    id_path = desc_signature.attributes["ids"][0]
    module = importlib.import_module(module_path)
    outer = inner = module
    for path in object_path.split("."):
        outer = inner
        inner = getattr(outer, path)
    if isinstance(outer, type):
        _, _, attr_name = object_path.rpartition(".")
        try:
            attr = {attr.name: attr for attr in inspect.classify_class_attrs(outer)}[
                attr_name
            ]
        except KeyError:
            return id_path
        if attr.defining_class is not outer:
            return ".".join(
                [
                    attr.defining_class.__module__,
                    attr.defining_class.__name__,
                    attr_name,
                ]
            )
    return id_path


def parse_rst(rst_string):
    parser = Parser()
    for name, class_ in []:  # Add custom directives here
        register_directive(name, class_)
    settings = get_default_settings(Parser)
    document = new_document("test", settings)
    parser.parse(rst_string, document)
    document = parser.document
    return document


def rebuild_document(document, node_mapping):
    for old_node, replacements in reversed(tuple(node_mapping.items())):
        new_nodes = []
        for is_vanilla, grouper in itertools.groupby(
            replacements, lambda x: isinstance(x, (ConsoleInput, ConsoleOutput))
        ):
            if is_vanilla:
                text = next(grouper).string
                for item in grouper:
                    if isinstance(item, ConsoleInput):
                        new_nodes.append(literal_block(text, text))
                        text = item.string
                    else:
                        text += item.string
                new_nodes.append(literal_block(text, text))
            else:
                for replacement in grouper:
                    new_nodes.extend(replacement.to_docutils())
        old_node.parent.replace(old_node, new_nodes)



import unittest

import uqbar.graphs
import uqbar.strings


class TestCase(unittest.TestCase):
    def test_functional(self):
        # instantiate them
        graph = uqbar.graphs.Graph(name="G")
        cluster_0 = uqbar.graphs.Graph(name="0", is_cluster=True)
        cluster_1 = uqbar.graphs.Graph(name="1", is_cluster=True)
        a0 = uqbar.graphs.Node(name="a0")
        a1 = uqbar.graphs.Node(name="a1")
        a2 = uqbar.graphs.Node(name="a2")
        a3 = uqbar.graphs.Node(name="a3")
        b0 = uqbar.graphs.Node(name="b0")
        b1 = uqbar.graphs.Node(name="b1")
        b2 = uqbar.graphs.Node(name="b2")
        b3 = uqbar.graphs.Node(name="b3")
        start = uqbar.graphs.Node(name="start")
        end = uqbar.graphs.Node(name="end")
        # aggregate them
        graph.extend([cluster_0, cluster_1, start, end])
        cluster_0.extend([a0, a1, a2, a3])
        cluster_1.extend([b0, b1, b2, b3])
        # attach them
        start.attach(a0)
        start.attach(b0)
        a0.attach(a1)
        a1.attach(a2)
        a1.attach(b3)
        a2.attach(a3)
        a3.attach(a0)
        a3.attach(end)
        b0.attach(b1)
        b1.attach(b2)
        b2.attach(b3)
        b2.attach(a3)
        b3.attach(end)
        # style them
        cluster_0.attributes["style"] = "filled"
        cluster_0.attributes["color"] = "lightgrey"
        cluster_0.attributes["label"] = "process #1"
        cluster_0.node_attributes["style"] = "filled"
        cluster_0.node_attributes["color"] = "white"
        cluster_1.attributes["color"] = "blue"
        cluster_1.attributes["label"] = "process #2"
        cluster_1.node_attributes["style"] = ("filled", "rounded")
        start.attributes["shape"] = "Mdiamond"
        end.attributes["shape"] = "Msquare"
        # format them
        assert format(graph) == repr(graph)
        assert format(graph, "graphviz") == uqbar.strings.normalize(
            """
            digraph G {
                subgraph cluster_0 {
                    graph [color=lightgrey,
                        label="process #1",
                        style=filled];
                    node [color=white,
                        style=filled];
                    a0;
                    a1;
                    a2;
                    a3;
                    a0 -> a1;
                    a1 -> a2;
                    a2 -> a3;
                    a3 -> a0;
                }
                subgraph cluster_1 {
                    graph [color=blue,
                        label="process #2"];
                    node [style="filled, rounded"];
                    b0;
                    b1;
                    b2;
                    b3;
                    b0 -> b1;
                    b1 -> b2;
                    b2 -> b3;
                }
                start [shape=Mdiamond];
                end [shape=Msquare];
                a1 -> b3;
                a3 -> end;
                b2 -> a3;
                b3 -> end;
                start -> a0;
                start -> b0;
            }
        """
        )

    def test___format___graphviz_01(self):
        graph = uqbar.graphs.Graph(name="G")
        cluster_foo = uqbar.graphs.Graph(name="foo", is_cluster=True)
        subgraph_bar = uqbar.graphs.Graph(name="bar")
        node_a = uqbar.graphs.Node(name="a")
        node_b = uqbar.graphs.Node(name="b")
        node_c = uqbar.graphs.Node(name="c")
        graph.extend([cluster_foo, subgraph_bar, node_a])
        cluster_foo.append(node_b)
        subgraph_bar.append(node_c)
        node_a.attach(node_b)
        node_a.attach(node_c)
        node_b.attach(node_c)
        assert format(graph, "graphviz") == uqbar.strings.normalize(
            """
        digraph G {
            subgraph cluster_foo {
                b;
            }
            subgraph bar {
                c;
            }
            a;
            b -> c;
            a -> b;
            a -> c;
        }
        """
        )
        cluster_foo.append(node_c)
        assert format(graph, "graphviz") == uqbar.strings.normalize(
            """
            digraph G {
                subgraph cluster_foo {
                    b;
                    c;
                    b -> c;
                }
                subgraph bar {
                }
                a;
                a -> b;
                a -> c;
            }
        """
        )
        subgraph_bar.extend([node_a, node_b, node_c])
        assert format(graph, "graphviz") == uqbar.strings.normalize(
            """
            digraph G {
                subgraph cluster_foo {
                }
                subgraph bar {
                    a;
                    b;
                    c;
                    a -> b;
                    a -> c;
                    b -> c;
                }
            }
        """
        )
        subgraph_bar.append(node_a)
        assert format(graph, "graphviz") == uqbar.strings.normalize(
            """
            digraph G {
                subgraph cluster_foo {
                }
                subgraph bar {
                    b;
                    c;
                    a;
                    b -> c;
                    a -> b;
                    a -> c;
                }
            }
        """
        )
        subgraph_bar.remove(node_a)
        assert format(graph, "graphviz") == uqbar.strings.normalize(
            """
            digraph G {
                subgraph cluster_foo {
                }
                subgraph bar {
                    b;
                    c;
                    b -> c;
                }
            }
        """
        )

    def test_initialized_with_attributes(self):
        graph = uqbar.graphs.Graph(
            name="g",
            attributes=dict(
                bgcolor="transparent",
                color="lightslategrey",
                dpi=72,
                fontname="Arial",
                outputorder="edgesfirst",
                overlap="prism",
                rankdir="LR",
                ranksep=1,
                splines="spline",
                style=("dotted", "rounded"),
            ),
            edge_attributes=dict(penwidth=2),
            node_attributes=dict(
                fontname="Arial", fontsize=12, penwidth=2, style=("filled", "rounded")
            ),
        )
        assert format(graph, "graphviz") == uqbar.strings.normalize(
            """
            digraph g {
                graph [bgcolor=transparent,
                    color=lightslategrey,
                    dpi=72,
                    fontname=Arial,
                    outputorder=edgesfirst,
                    overlap=prism,
                    rankdir=LR,
                    ranksep=1,
                    splines=spline,
                    style="dotted, rounded"];
                node [fontname=Arial,
                    fontsize=12,
                    penwidth=2,
                    style="filled, rounded"];
                edge [penwidth=2];
            }
        """
        )



import os
import pathlib
import shutil
import sys

import pytest

import uqbar.apis
from uqbar.strings import normalize


@pytest.fixture
def test_path():
    test_path = pathlib.Path(__file__).parent
    docs_path = test_path / "docs"
    if str(test_path) not in sys.path:
        sys.path.insert(0, str(test_path))
    if docs_path.exists():
        shutil.rmtree(str(docs_path))
    yield test_path
    if docs_path.exists():
        shutil.rmtree(str(docs_path))


def test_str_01(test_path):
    builder = uqbar.apis.APIBuilder(
        [test_path / "fake_package"],
        test_path / "docs",
        member_documenter_classes=[
            uqbar.apis.FunctionDocumenter,
            uqbar.apis.SummarizingClassDocumenter,
        ],
        module_documenter_class=uqbar.apis.SummarizingModuleDocumenter,
        root_documenter_class=uqbar.apis.SummarizingRootDocumenter,
    )
    builder()
    paths = sorted((test_path / "docs").rglob("*"))
    paths = [
        str(path.relative_to(test_path)).replace(os.path.sep, "/") for path in paths
    ]
    assert paths == [
        "docs/fake_package",
        "docs/fake_package/empty_module.rst",
        "docs/fake_package/empty_package",
        "docs/fake_package/empty_package/empty.rst",
        "docs/fake_package/empty_package/index.rst",
        "docs/fake_package/enums.rst",
        "docs/fake_package/index.rst",
        "docs/fake_package/module.rst",
        "docs/fake_package/multi",
        "docs/fake_package/multi/index.rst",
        "docs/fake_package/multi/one.rst",
        "docs/fake_package/multi/two.rst",
        "docs/index.rst",
    ]
    base_path = test_path / "docs" / "fake_package"
    with (base_path / ".." / "index.rst").open() as file_pointer:
        assert normalize(file_pointer.read()) == normalize(
            """
            API
            ===

            .. toctree::
               :hidden:

               fake_package/index

            .. raw:: html

               <hr/>

            .. rubric:: :ref:`fake_package <fake-package>`
               :class: section-header

            .. raw:: html

               <hr/>

            .. rubric:: :ref:`fake_package.empty_module <fake-package--empty-module>`
               :class: section-header

            .. raw:: html

               <hr/>

            .. rubric:: :ref:`fake_package.empty_package <fake-package--empty-package>`
               :class: section-header

            .. raw:: html

               <hr/>

            .. rubric:: :ref:`fake_package.empty_package.empty <fake-package--empty-package--empty>`
               :class: section-header

            .. raw:: html

               <hr/>

            .. rubric:: :ref:`fake_package.enums <fake-package--enums>`
               :class: section-header

            .. raw:: html

               <hr/>

            .. rubric:: Enumerations
               :class: subsection-header

            .. autosummary::
               :nosignatures:

               ~fake_package.enums.FakeEnum

            .. raw:: html

               <hr/>

            .. rubric:: :ref:`fake_package.module <fake-package--module>`
               :class: section-header

            .. raw:: html

               <hr/>

            .. rubric:: Classes
               :class: subsection-header

            .. autosummary::
               :nosignatures:

               ~fake_package.module.ChildClass
               ~fake_package.module.PublicClass

            .. raw:: html

               <hr/>

            .. rubric:: Functions
               :class: subsection-header

            .. autosummary::
               :nosignatures:

               ~fake_package.module.public_function

            .. raw:: html

               <hr/>

            .. rubric:: :ref:`fake_package.multi <fake-package--multi>`
               :class: section-header

            .. raw:: html

               <hr/>

            .. rubric:: Classes
               :class: subsection-header

            .. autosummary::
               :nosignatures:

               ~fake_package.multi.PublicClass

            .. raw:: html

               <hr/>

            .. rubric:: Functions
               :class: subsection-header

            .. autosummary::
               :nosignatures:

               ~fake_package.multi.public_function

            .. raw:: html

               <hr/>

            .. rubric:: :ref:`fake_package.multi.one <fake-package--multi--one>`
               :class: section-header

            .. raw:: html

               <hr/>

            .. rubric:: Classes
               :class: subsection-header

            .. autosummary::
               :nosignatures:

               ~fake_package.multi.one.PublicClass

            .. raw:: html

               <hr/>

            .. rubric:: Functions
               :class: subsection-header

            .. autosummary::
               :nosignatures:

               ~fake_package.multi.one.public_function

            .. raw:: html

               <hr/>

            .. rubric:: :ref:`fake_package.multi.two <fake-package--multi--two>`
               :class: section-header

            .. raw:: html

               <hr/>

            .. rubric:: Classes
               :class: subsection-header

            .. autosummary::
               :nosignatures:

               ~fake_package.multi.two.PublicClass

            .. raw:: html

               <hr/>

            .. rubric:: Functions
               :class: subsection-header

            .. autosummary::
               :nosignatures:

               ~fake_package.multi.two.public_function
            """
        )
    with (base_path / "index.rst").open() as file_pointer:
        assert normalize(file_pointer.read()) == normalize(
            """
            .. _fake-package:

            fake_package
            ============

            .. automodule:: fake_package

            .. currentmodule:: fake_package

            .. container:: svg-container

               .. inheritance-diagram:: fake_package
                  :lineage: fake_package

            .. raw:: html

               <hr/>

            .. rubric:: Subpackages
               :class: section-header

            .. toctree::
               :hidden:

               empty_module
               empty_package/index
               enums
               module
               multi/index

            .. autosummary::
               :nosignatures:

               empty_module
               empty_package
               enums
               module
               multi
            """
        )



"""
Tools for building Graphviz graphs.
"""

from .attrs import Attributes
from .core import Attachable, Edge, Graph, Node
from .graphers import Grapher
from .html import HRule, LineBreak, Table, TableCell, TableRow, Text, VRule
from .records import RecordField, RecordGroup

__all__ = [
    "Attachable",
    "Attributes",
    "Edge",
    "Graph",
    "Grapher",
    "HRule",
    "LineBreak",
    "Node",
    "RecordField",
    "RecordGroup",
    "Table",
    "TableCell",
    "TableRow",
    "Text",
    "VRule",
]



from uqbar.objects import get_hash


class MyObject:
    def __init__(self, arg1, arg2, *var_args, foo=None, bar=None, **kwargs):
        self.arg1 = arg1
        self.arg2 = arg2
        self.var_args = var_args
        self.foo = foo
        self.bar = bar
        self.kwargs = kwargs


def test_objects_get_hash_01():
    object_a = MyObject("a", "b", "c", "d", foo="x", quux=["y", "z"])
    object_b = MyObject("a", "b", "c", "d", foo="x", quux=["y", "z"])
    object_c = MyObject("x", "y", bar=set([1, 2, 3]))
    assert get_hash(object_a) == get_hash(object_b)
    assert get_hash(object_a) != get_hash(object_c)
    assert get_hash(object_b) != get_hash(object_c)



"""
Uqbar Sphinx styling extension.

- Adds additional information to autodoc entries.
- Adds additional CSS.

Install by adding ``'uqbar.sphinx.style'`` to the ``extensions`` list in your
Sphinx configuration.
"""

import importlib
import inspect
import pathlib
from typing import Any, Dict

from docutils import nodes
from sphinx import addnodes  # type: ignore
from sphinx.locale import _
from sphinx.util import logging

import uqbar.io

logger = logging.getLogger(__name__)


def handle_class(signature_node, module, object_name, cache):
    """
    Styles ``autoclass`` entries.

    Adds ``abstract`` prefix to abstract classes.
    """
    class_ = getattr(module, object_name, None)
    if class_ is None:
        return
    if class_ not in cache:
        cache[class_] = {}
        attributes = inspect.classify_class_attrs(class_)
        for attribute in attributes:
            cache[class_][attribute.name] = attribute
    if inspect.isabstract(class_):
        emphasis = nodes.emphasis("abstract ", "abstract ", classes=["property"])
        signature_node.insert(0, emphasis)


def handle_method(signature_node, module, object_name, cache):
    """
    Styles ``automethod`` entries.

    Adds link to originating class for inherited methods.
    """
    *class_names, attr_name = object_name.split(".")  # Handle nested classes
    class_ = module
    for class_name in class_names:
        class_ = getattr(class_, class_name, None)
        if class_ is None:
            return
    try:
        inspected_attr = cache[class_][attr_name]
        defining_class = inspected_attr.defining_class
    except KeyError:
        # TODO: This is a hack to handle bad interaction between enum and inspect
        defining_class = class_
    if defining_class is not class_:
        reftarget = "{}.{}".format(defining_class.__module__, defining_class.__name__)
        xref_node = addnodes.pending_xref(
            "", refdomain="py", refexplicit=True, reftype="class", reftarget=reftarget
        )
        name_node = nodes.literal(
            "", "{}".format(defining_class.__name__), classes=["descclassname"]
        )
        xref_node.append(name_node)
        desc_annotation = list(signature_node.findall(addnodes.desc_annotation))
        index = len(desc_annotation)
        class_annotation = addnodes.desc_addname()
        class_annotation.extend([nodes.Text("("), xref_node, nodes.Text(").")])
        class_annotation["xml:space"] = "preserve"
        signature_node.insert(index, class_annotation)
    else:
        is_overridden = False
        for class_ in defining_class.__mro__[1:]:
            if hasattr(class_, attr_name):
                is_overridden = True
        if is_overridden:
            emphasis = nodes.emphasis(
                "overridden ", "overridden ", classes=["property"]
            )
            signature_node.insert(0, emphasis)


def on_doctree_read(app, document) -> None:
    """
    Hooks into Sphinx's ``doctree-read`` event.
    """
    cache: Dict[type, Dict[str, object]] = {}
    for desc_node in document.findall(addnodes.desc):
        if desc_node.get("domain") != "py":
            continue
        signature_node = list(desc_node.findall(addnodes.desc_signature))[0]
        module_name = signature_node.get("module")
        object_name = signature_node.get("fullname")
        object_type = desc_node.get("objtype")
        module = importlib.import_module(module_name)
        if object_type == "class":
            handle_class(signature_node, module, object_name, cache)
        elif object_type in ("method", "attribute", "staticmethod", "classmethod"):
            handle_method(signature_node, module, object_name, cache)


def on_builder_inited(app):
    """
    Hooks into Sphinx's ``builder-inited`` event.

    Used for copying over CSS files to theme directory.
    """
    local_css_path = pathlib.Path(__file__).parent / "uqbar.css"
    theme_css_path = (
        pathlib.Path(app.srcdir) / app.config.html_static_path[0] / "uqbar.css"
    )
    with local_css_path.open("r") as file_pointer:
        local_css_contents = file_pointer.read()
    uqbar.io.write(local_css_contents, theme_css_path)


# Make definition list terms linkable


def visit_classifier(self, node):
    self.body.append(self.starttag(node, "span", "", CLASS="classifier"))


def depart_classifier(self, node):
    self.body.append("</span>")
    next_node = node.next_node(descend=False, siblings=True)
    if not isinstance(next_node, nodes.classifier):
        self.body.append("</dt>")


def visit_definition(self, node):
    self.body.append(self.starttag(node, "dd", ""))


def depart_definition(self, node):
    self.body.append("</dd>\n")


def visit_term(self, node):
    self.body.append(self.starttag(node, "dt", ""))


def depart_term(self, node):
    next_node = node.next_node(descend=False, siblings=True)
    if isinstance(next_node, nodes.classifier):
        pass
    else:
        self.add_permalink_ref(node, _("Permalink to this term"))
        self.body.append("</dt>")


# Setup


def setup(app) -> Dict[str, Any]:
    """
    Sets up Sphinx extension.
    """
    app.connect("doctree-read", on_doctree_read)
    app.connect("builder-inited", on_builder_inited)
    app.add_css_file("uqbar.css")
    app.add_node(
        nodes.classifier, override=True, html=(visit_classifier, depart_classifier)
    )
    app.add_node(
        nodes.definition, override=True, html=(visit_definition, depart_definition)
    )
    app.add_node(nodes.term, override=True, html=(visit_term, depart_term))
    return {
        "version": uqbar.__version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



import uqbar.objects
import uqbar.strings


class MyObject:
    def __init__(self, arg1, arg2, *var_args, foo=None, bar=None, **kwargs):
        self.arg1 = arg1
        self.arg2 = arg2
        self.var_args = var_args
        self.foo = foo
        self.bar = bar
        self.kwargs = kwargs

    def __repr__(self):
        return uqbar.objects.get_repr(self)


def test_objects_new_01():
    object_a = MyObject("a", "b", "c", "d", foo="x", quux=["y", "z"])
    object_b = uqbar.objects.new(
        object_a, arg2=MyObject("x", "y", bar=[1, 2, 3]), foo="FOO"
    )
    # object A is unchanged
    assert repr(object_a) == uqbar.strings.normalize(
        """
        MyObject(
            'a',
            'b',
            'c',
            'd',
            foo='x',
            quux=['y', 'z'],
        )
        """
    )
    assert repr(object_b) == uqbar.strings.normalize(
        """
        MyObject(
            'a',
            MyObject(
                'x',
                'y',
                bar=[1, 2, 3],
            ),
            'c',
            'd',
            foo='FOO',
            quux=['y', 'z'],
        )
        """
    )
    object_c = uqbar.objects.new(
        object_b, arg1="new a", arg2__foo="new FOO", arg2__bar=[4, 5, 6, 7]
    )
    # object B is unchanged
    assert repr(object_b) == uqbar.strings.normalize(
        """
        MyObject(
            'a',
            MyObject(
                'x',
                'y',
                bar=[1, 2, 3],
            ),
            'c',
            'd',
            foo='FOO',
            quux=['y', 'z'],
        )
        """
    )
    assert repr(object_c) == uqbar.strings.normalize(
        """
        MyObject(
            'new a',
            MyObject(
                'x',
                'y',
                bar=[4, 5, 6, 7],
                foo='new FOO',
            ),
            'c',
            'd',
            foo='FOO',
            quux=['y', 'z'],
        )
        """
    )



"""
Tools for auto-generating API documention.
"""

import importlib
import pathlib

from .builders import APIBuilder
from .documenters import (
    ClassDocumenter,
    FunctionDocumenter,
    MemberDocumenter,
    ModuleDocumenter,
    RootDocumenter,
)
from .graphs import InheritanceGraph
from .nodes import ModuleNode, PackageNode
from .summarizers import (
    ImmaterialClassDocumenter,
    ImmaterialModuleDocumenter,
    SummarizingClassDocumenter,
    SummarizingModuleDocumenter,
    SummarizingRootDocumenter,
)


def collect_source_paths(source_paths, recurse_subpackages=True):
    visited_paths = set()
    path_stack = []
    for source_path in source_paths:
        source_path = pathlib.Path(source_path).resolve().absolute()
        assert source_path.exists(), source_path
        path_stack.append(source_path)
    while path_stack:
        current_path = path_stack.pop()
        if current_path in visited_paths:
            continue
        if current_path.is_dir():
            if not (current_path / "__init__.py").exists():
                continue
            if recurse_subpackages:
                for path in current_path.iterdir():
                    if path.is_dir() and (path / "__init__.py").exists():
                        path_stack.append(path)
                    elif path.suffix in (".py", ".pyx"):
                        visited_paths.add(path)
            else:
                visited_paths.add(current_path / "__init__.py")
                for path in current_path.iterdir():
                    if path.is_dir():
                        continue
                    # Permit nominative modules
                    if path.suffix in (".py", ".pyx"):
                        package_path = source_path_to_package_path(path)
                        module = importlib.import_module(package_path)
                        if hasattr(module, path.with_suffix("").name):
                            visited_paths.add(path)
        elif current_path.suffix in (".py", ".pyx"):
            visited_paths.add(current_path)
    return sorted(visited_paths)


def source_path_to_package_path(path):
    path = pathlib.Path(path)
    if not (path.parent / "__init__.py").exists():
        return path.with_suffix("").name
    root = path
    while (root.parent / "__init__.py").exists():
        root = root.parent
    path = path.with_suffix("")
    if path == root:
        return path.name
    parts = (root.name,) + path.relative_to(root).parts
    if parts[-1] == "__init__":
        parts = parts[:-1]
    return ".".join(parts)


__all__ = [
    "APIBuilder",
    "ClassDocumenter",
    "FunctionDocumenter",
    "ImmaterialClassDocumenter",
    "ImmaterialModuleDocumenter",
    "InheritanceGraph",
    "MemberDocumenter",
    "ModuleDocumenter",
    "ModuleNode",
    "PackageNode",
    "RootDocumenter",
    "SummarizingClassDocumenter",
    "SummarizingModuleDocumenter",
    "SummarizingRootDocumenter",
    "collect_source_paths",
    "source_path_to_package_path",
]



import pathlib
import sys

import pytest

import uqbar.apis
from uqbar.strings import normalize


@pytest.fixture
def test_path():
    test_path = pathlib.Path(__file__).parent
    if str(test_path) not in sys.path:
        sys.path.insert(0, str(test_path))


def test_str_01(test_path):
    documenter = uqbar.apis.SummarizingModuleDocumenter("fake_package.module")
    assert normalize(str(documenter)) == normalize(
        """
        .. _fake-package--module:

        module
        ======

        .. automodule:: fake_package.module

        .. currentmodule:: fake_package.module

        .. container:: svg-container

           .. inheritance-diagram:: fake_package
              :lineage: fake_package.module

        .. raw:: html

           <hr/>

        .. rubric:: Classes
           :class: section-header

        .. autosummary::
           :nosignatures:

           ~ChildClass
           ~PublicClass

        .. autoclass:: ChildClass
           :members:
           :undoc-members:

        .. autoclass:: PublicClass
           :members:
           :undoc-members:

        .. raw:: html

           <hr/>

        .. rubric:: Functions
           :class: section-header

        .. autosummary::
           :nosignatures:

           ~public_function

        .. autofunction:: public_function
        """
    )


def test_str_02(test_path):
    documenter = uqbar.apis.SummarizingModuleDocumenter(
        "fake_package.module", document_private_members=True
    )
    assert normalize(str(documenter)) == normalize(
        """
        .. _fake-package--module:

        module
        ======

        .. automodule:: fake_package.module

        .. currentmodule:: fake_package.module

        .. container:: svg-container

           .. inheritance-diagram:: fake_package
              :lineage: fake_package.module

        .. raw:: html

           <hr/>

        .. rubric:: Classes
           :class: section-header

        .. autosummary::
           :nosignatures:

           ~ChildClass
           ~PublicClass
           ~_PrivateClass

        .. autoclass:: ChildClass
           :members:
           :undoc-members:

        .. autoclass:: PublicClass
           :members:
           :undoc-members:

        .. autoclass:: _PrivateClass
           :members:
           :undoc-members:

        .. raw:: html

           <hr/>

        .. rubric:: Functions
           :class: section-header

        .. autosummary::
           :nosignatures:

           ~_private_function
           ~public_function

        .. autofunction:: _private_function

        .. autofunction:: public_function
        """
    )


def test_str_03(test_path):
    documenter = uqbar.apis.SummarizingModuleDocumenter(
        "fake_package.module",
        member_documenter_classes=[
            uqbar.apis.FunctionDocumenter,
            uqbar.apis.SummarizingClassDocumenter,
        ],
    )
    assert normalize(str(documenter)) == normalize(
        """
        .. _fake-package--module:

        module
        ======

        .. automodule:: fake_package.module

        .. currentmodule:: fake_package.module

        .. container:: svg-container

           .. inheritance-diagram:: fake_package
              :lineage: fake_package.module

        .. raw:: html

           <hr/>

        .. rubric:: Classes
           :class: section-header

        .. autosummary::
           :nosignatures:

           ~ChildClass
           ~PublicClass

        .. autoclass:: ChildClass

           .. raw:: html

              <hr/>

           .. rubric:: Attributes Summary
              :class: class-header

           .. autosummary::
              :nosignatures:

              inheritable_method
              new_method

           .. raw:: html

              <hr/>

           .. rubric:: Special methods
              :class: class-header

           .. container:: inherited

              .. automethod:: ChildClass.__str__

           .. raw:: html

              <hr/>

           .. rubric:: Methods
              :class: class-header

           .. automethod:: ChildClass.inheritable_method

           .. container:: inherited

              .. automethod:: ChildClass.method

           .. automethod:: ChildClass.new_method

           .. container:: inherited

              .. automethod:: ChildClass.other_method

           .. raw:: html

              <hr/>

           .. rubric:: Class & static methods
              :class: class-header

           .. container:: inherited

              .. automethod:: ChildClass.class_method

           .. container:: inherited

              .. automethod:: ChildClass.static_method

           .. raw:: html

              <hr/>

           .. rubric:: Read/write properties
              :class: class-header

           .. container:: inherited

              .. autoattribute:: ChildClass.read_write_property

           .. raw:: html

              <hr/>

           .. rubric:: Read-only properties
              :class: class-header

           .. container:: inherited

              .. autoattribute:: ChildClass.read_only_property

        .. autoclass:: PublicClass

           .. raw:: html

              <hr/>

           .. rubric:: Attributes Summary
              :class: class-header

           .. autosummary::
              :nosignatures:

              __str__
              class_method
              inheritable_method
              method
              other_method
              read_only_property
              read_write_property
              static_method

           .. raw:: html

              <hr/>

           .. rubric:: Special methods
              :class: class-header

           .. automethod:: PublicClass.__str__

           .. raw:: html

              <hr/>

           .. rubric:: Methods
              :class: class-header

           .. automethod:: PublicClass.inheritable_method

           .. automethod:: PublicClass.method

           .. automethod:: PublicClass.other_method

           .. raw:: html

              <hr/>

           .. rubric:: Class & static methods
              :class: class-header

           .. automethod:: PublicClass.class_method

           .. automethod:: PublicClass.static_method

           .. raw:: html

              <hr/>

           .. rubric:: Read/write properties
              :class: class-header

           .. autoattribute:: PublicClass.read_write_property

           .. raw:: html

              <hr/>

           .. rubric:: Read-only properties
              :class: class-header

           .. autoattribute:: PublicClass.read_only_property

        .. raw:: html

           <hr/>

        .. rubric:: Functions
           :class: section-header

        .. autosummary::
           :nosignatures:

           ~public_function

        .. autofunction:: public_function
        """
    )
